"""
This module creates clustered subsets of features described in the paper Clustered Feature Importance (Presentation
Slides) by Dr. Marcos Lopez de Prado. https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3517595 and is also explained
in the book Machine Learning for Asset Managers Snippet 6.5.2 page 84.
"""

#Imports
import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage, fcluster
from statsmodels.regression.linear_model import OLS

from famlafl.clustering.onc import get_onc_clusters
from famlafl.codependence.codependence_matrix import get_dependence_matrix, get_distance_matrix


# pylint: disable=invalid-name
def get_feature_clusters(X: pd.DataFrame, dependence_metric: str, distance_metric: str = None,
                         linkage_method: str = None, n_clusters: int = None, critical_threshold: float = 0.0) -> list:
    """
    Machine Learning for Asset Managers
    Snippet 6.5.2.1 , page 85. Step 1: Features Clustering

    Gets clustered features subsets from the given set of features.

    :param X: (pd.DataFrame) Dataframe of features.
    :param dependence_metric: (str) Method to be use for generating dependence_matrix, either 'linear' or
                              'information_variation' or 'mutual_information' or 'distance_correlation'.
    :param distance_metric: (str) The distance operator to be used for generating the distance matrix. The methods that
                            can be applied are: 'angular', 'squared_angular', 'absolute_angular'. Set it to None if the
                            feature are to be generated as it is by the ONC algorithm.
    :param linkage_method: (str) Method of linkage to be used for clustering. Methods include: 'single', 'ward',
                           'complete', 'average', 'weighted', and 'centroid'. Set it to None if the feature are to
                           be generated as it is by the ONC algorithm.
    :param n_clusters: (int) Number of clusters to form. Must be less the total number of features. If None then it
                       returns optimal number of clusters decided by the ONC Algorithm.
    :param critical_threshold: (float) Threshold for determining low silhouette score in the dataset. It can any real number
                                in [-1,+1], default is 0 which means any feature that has a silhouette score below 0 will be
                                indentified as having low silhouette and hence requied transformation will be appiled to for
                                for correction of the same.
    :return: (list) Feature subsets.
    """
    # Checking if dataset contains features low silhouette
    X = _check_for_low_silhouette_scores(X, critical_threshold)

    # Get the dependence matrix
    if dependence_metric != 'linear':
        dep_matrix = get_dependence_matrix(X, dependence_method=dependence_metric)
    else:
        dep_matrix = X.corr()

    if n_clusters is None and (distance_metric is None or linkage_method is None):
        return list(get_onc_clusters(dep_matrix.fillna(0))[1].values())  # Get optimal number of clusters
    if distance_metric is not None and (linkage_method is not None and n_clusters is None):
        n_clusters = len(get_onc_clusters(dep_matrix.fillna(0))[1])
    if n_clusters >= len(X.columns):  # Check if number of clusters exceeds number of features
        raise ValueError('Number of clusters must be less than the number of features')

    # Apply distance operator on the dependence matrix
    dist_matrix = get_distance_matrix(dep_matrix, distance_metric=distance_metric)

    # Get the linkage
    link = linkage(squareform(dist_matrix), method=linkage_method)
    clusters = fcluster(link, t=n_clusters, criterion='maxclust')
    clustered_subsets = [[f for c, f in zip(clusters, X.columns) if c == ci] for ci in range(1, n_clusters + 1)]

    return clustered_subsets


def _cluster_transformation(X: pd.DataFrame, clusters: dict, feats_to_transform: list) -> pd.DataFrame:
    """
    Machine Learning for Asset Managers
    Snippet 6.5.2.1 , page 85. Step 1: Features Clustering (last paragraph)

    Transforms a dataset to reduce the multicollinearity of the system by replacing the original feature with
    the residual from regression.

    :param X: (pd.DataFrame) Dataframe of features.
    :param clusters: (dict) Clusters generated by ONC algorithm.
    :param feats_to_transform: (list) Features that have low silhouette score and to be transformed.
    :return: (pd.DataFrame) Transformed features.
    """
    for feat in feats_to_transform:
        for i, j in clusters.items():

            if feat in j:  # Selecting the cluster that contains the feature
                exog = sm.add_constant(X.drop(j, axis=1)).values
                endog = X[feat].values
                ols = OLS(endog, exog).fit()

                if ols.df_model < (exog.shape[1]-1):
                    # Degree of freedom is low
                    new_exog = _combine_features(X, clusters, i)
                    # Run the regression again on the new exog
                    ols = OLS(endog, new_exog.reshape(exog.shape[0], -1)).fit()
                    X[feat] = ols.resid
                else:
                    X[feat] = ols.resid

    return X


def _combine_features(X, clusters, exclude_key) -> np.array:
    """
    Combines features of each cluster linearly by following a minimum variance weighting scheme.
    The Minimum Variance weights are calculated without constraints, other than the weights sum to one.

    :param X: (pd.DataFrame) Dataframe of features.
    :param clusters: (dict) Clusters generated by ONC algorithm.
    :param exclude_key: (int) Key of the cluster which is to be excluded.
    :return: (np.array) Combined features for each cluster.
    """

    new_exog = []
    for i, cluster in clusters.items():

        if i != exclude_key:
            subset = X[cluster]
            cov_matx = subset.cov()  # Covariance matrix of the cluster
            eye_vec = np.array(cov_matx.shape[1]*[1], float)
            try:
                numerator = np.dot(np.linalg.inv(cov_matx), eye_vec)
                denominator = np.dot(eye_vec, numerator)
                # Minimum variance weighting
                wghts = numerator/denominator
            except np.linalg.LinAlgError:
                # A singular matrix so giving each component equal weight
                wghts = np.ones(subset.shape[1]) * (1/subset.shape[1])
            new_exog.append(((subset*wghts).sum(1)).values)

    return np.array(new_exog)


def _check_for_low_silhouette_scores(X: pd.DataFrame, critical_threshold: float = 0.0) -> pd.DataFrame:
    """
    Machine Learning for Asset Managers
    Snippet 6.5.2.1 , page 85. Step 1: Features Clustering (last paragraph)

    Checks where the dataset contains features low silhouette due one feature being a combination of
    multiple features across clusters. This is a problem, because ONC cannot assign one feature to multiple
    clusters and it needs a transformation.

    :param X: (pd.DataFrame) Dataframe of features.
    :param critical_threshold: (float) Threshold for determining low silhouette score.
    :return: (pd.DataFrame) Dataframe of features.
    """
    _, clstrs, silh = get_onc_clusters(X.corr())
    low_silh_feat = silh[silh < critical_threshold].index
    if len(low_silh_feat) > 0:
        print(f'{len(low_silh_feat)} feature/s found with low silhouette score {low_silh_feat}. Returning the transformed dataset')

        # Returning the transformed dataset
        return _cluster_transformation(X, clstrs, low_silh_feat)

    print('No feature/s found with low silhouette score. All features belongs to its respective clusters')

    return X



# pylint: disable=missing-module-docstring
import sys
import pandas as pd
import numpy as np
import cvxpy as cp


class OLPS:
    """
    Online Portfolio Selection is an algorithmic trading strategy that sequentially allocates
    capital among a group of assets to maximize the final returns of the investment.

    Traditional theories for portfolio selection, such as Markowitz’s Modern Portfolio Theory,
    optimize the balance between the portfolio's risks and returns. However, OLPS is founded on
    the capital growth theory, which solely focuses on maximizing the returns of the current
    portfolio.

    Through these walkthroughs of different portfolio selection strategies, we hope to introduce
    a set of different selection tools available for everyone. Most of the works will be based on
    Dr. Bin Li and Dr. Steven Hoi’s book, Online Portfolio Selection: Principles and Algorithms,
    and further recent papers will be implemented to assist the development and understanding of
    these unique portfolio selection strategies.

    OLPS is the parent class for all resulting Online Portfolio Selection Strategies.

    This class broadly defines all variables and allocates a set of weights for a certain strategy.

    Upon weights allocation the possible outputs are:

    - ``self.weights`` (np.array) Final portfolio weights prediction.

    - ``self.all_weights`` (pd.DataFrame) Portfolio weights for the time period.

    - ``self.asset_name`` (list) Name of assets.

    - ``self.number_of_assets`` (int) Number of assets.

    - ``self.time`` (datetime) Time index of the given data.

    - ``self.length_of_time`` (int) Number of time periods.

    - ``self.relative_return`` (np.array) Relative returns of the assets.

    - ``self.portfolio_return`` (pd.DataFrame) Cumulative portfolio returns over time.

    - ``self.asset_prices`` (pd.DataFrame) Historical asset prices (daily close).
    """

    def __init__(self):
        self.weights = None  # (np.array) Final portfolio weights prediction.
        self.all_weights = None  # (pd.DataFrame) Portfolio weights for the time period.
        self.asset_name = None  # (list) Name of assets.
        self.number_of_assets = None  # (int) Number of assets.
        self.time = None  # (datetime) Time index of the given data.
        self.length_of_time = None  # (int) Number of time periods.
        self.relative_return = None  # (np.array) Relative returns of the assets.
        self.portfolio_return = None  # (pd.DataFrame) Cumulative portfolio returns over time.
        self.asset_prices = None  # (pd.DataFrame) Historical asset prices (daily close).

    def allocate(self, asset_prices, weights=None, resample_by=None, verbose=False):
        """
        Allocates weight according to a set of update rules.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user. If None, weights
                                                     will default to uniform weights.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        :param verbose: (bool) Prints progress bar if true.
        """
        # Check to ensure inputs are correct.
        self._check_asset(asset_prices, weights)

        # Initialize all variables.
        self._initialize(asset_prices, weights, resample_by)

        # Iterate through data and calculate weights.
        self._run(weights, verbose)

        # Round weights and drop values that are less than the given threshold.
        self._round_weights(threshold=1e-6)

        # Calculate portfolio returns based on weights calculated from the run method.
        self._calculate_portfolio_returns(self.all_weights, self.relative_return)

        # Convert everything to dataframe to make the information presentable.
        self._conversion()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        # Resample asset.
        if resample_by is not None:
            asset_prices = asset_prices.resample(resample_by).last()

        # Set asset names.
        self.asset_name = asset_prices.columns

        # Set time.
        self.time = asset_prices.index

        # Calculate number of assets.
        self.number_of_assets = self.asset_name.size

        # Calculate number of time.
        self.length_of_time = self.time.size

        # Calculate relative returns and final relative returns.
        self.relative_return = self._calculate_relative_return(asset_prices)

        # Set initial weights.
        self.weights = weights

        # Set all_weights.
        self.all_weights = np.zeros((self.length_of_time + 1, self.number_of_assets))

        # Set portfolio returns.
        self.portfolio_return = np.zeros((self.length_of_time, 1))

        # Pass dataframe to speed up process for universal portfolio.
        self.asset_prices = asset_prices

    def _run(self, weights, verbose):
        """
        Runs the algorithm by iterating through the given data.

        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param verbose: (bool) Prints progress bar if true.
        """
        # Set initial weights.
        self.weights = self._first_weight(weights)

        # Set initial_all weights to be the first given weight.
        self.all_weights[0] = self.weights

        # Run the algorithm for the rest of data from time 1.
        for time in range(self.length_of_time):
            # Update weights.
            self.weights = self._update_weight(time)
            self.all_weights[time + 1] = self.weights
            # Print progress bar.
            if verbose:
                self._print_progress(time + 1, prefix='Progress:', suffix='Complete')

        # Remove final prediction as that information is stored in self.weights.
        self.all_weights = self.all_weights[:-1]

    def _first_weight(self, weights):
        """
        Returns the first weight of the given portfolio. If the first weight is not given, initialize weights to
        uniform weights.

        :param weights: (list/np.array/pd.Dataframe) Initial weights set by the user.
        :return: (np.array) First portfolio weight.
        """
        # If no weights are given, return uniform weights.
        if weights is None:
            weights = self._uniform_weight()
        return weights

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Identical weights are returned for OLPS.
        """
        # Weights do not change for this class.
        new_weights = self.all_weights[time]
        return new_weights

    def _calculate_portfolio_returns(self, all_weights, relative_return):
        """
        Calculates cumulative portfolio returns.

        :param all_weights: (np.array) Portfolio weights for the time period.
        :param relative_return: (np.array) Relative returns of the assets.
        """
        # Take the dot product of the relative returns and transpose of all_weights.
        # Diagonal of the resulting matrix will be the returns for each week.
        # Calculate returns by taking the cumulative product.
        self.portfolio_return = np.diagonal(np.dot(relative_return, all_weights.T)).cumprod()

    def _conversion(self):
        """
        Converts the given np.array to pd.Dataframe.
        """
        # Convert all_weights.
        self.all_weights = pd.DataFrame(self.all_weights, index=self.time, columns=self.asset_name)

        # Convert portfolio_return.
        self.portfolio_return = pd.DataFrame(self.portfolio_return, index=self.time, columns=["Returns"])

    def _optimize(self, optimize_array, solver=cp.SCS):
        """
        Calculates weights that maximize returns over the given array.

        :param optimize_array: (np.array) Relative returns of the assets for a given time period.
        :param solver: (cp.solver) Solver for cvxpy
        :return: (np.array) Weights that maximize the returns for the given array.
        """

        # Initialize weights for the optimization problem.
        weights = cp.Variable(self.number_of_assets)

        # Use cp.log and cp.sum to make the cost function a convex function.
        # Multiplying continuous returns equates to summing over the log returns.
        portfolio_return = cp.sum(cp.log(optimize_array @ weights))

        # Optimization objective and constraints.
        allocation_objective = cp.Maximize(portfolio_return)
        allocation_constraints = [cp.sum(weights) == 1, cp.min(weights) >= 0]

        # Define and solve the problem.
        problem = cp.Problem(objective=allocation_objective, constraints=allocation_constraints)

        # Solve and return the resulting weights.
        problem.solve(warm_start=True, solver=solver)
        return weights.value

    def _round_weights(self, threshold=1e-6):
        """
        Drops weights that are below a certain threshold.

        :param threshold: (float) Drop all values below this threshold.
        """
        # Set all values below the threshold to 0.
        new_all_weights = np.where(self.all_weights < threshold, 0, self.all_weights)

        # Adjust weights to have the weights sum to 1.
        new_all_weights = np.apply_along_axis(lambda x: x / np.sum(x), 1, new_all_weights)
        self.all_weights = new_all_weights

    def _uniform_weight(self):
        """
        Returns a uniform weight of assets.

        :return: (np.array) Uniform weights (1/n, 1/n, 1/n ...).
        """
        # Divide by number of assets after creating numpy arrays of one.
        uni_weight = np.ones(self.number_of_assets) / self.number_of_assets
        return uni_weight

    def _print_progress(self, iteration, prefix='', suffix='', decimals=1, bar_length=50):
        # pylint: disable=expression-not-assigned
        """
        Calls in a loop to create a terminal progress bar.
        https://gist.github.com/aubricus/f91fb55dc6ba5557fbab06119420dd6a

        :param iteration: (int) Current iteration.
        :param prefix: (str) Prefix string.
        :param suffix: (str) Suffix string.
        :param decimals: (int) Positive number of decimals in percent completed.
        :param bar_length: (int) Character length of the bar.
        """
        str_format = "{0:." + str(decimals) + "f}"
        # Calculate the percent completed.
        percents = str_format.format(100 * (iteration / float(self.length_of_time)))
        # Calculate the length of bar.
        filled_length = int(round(bar_length * iteration / float(self.length_of_time)))
        # Fill the bar.
        block = '█' * filled_length + '-' * (bar_length - filled_length)
        # Print new line.
        sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, block, percents, '%', suffix)),

        if iteration == self.length_of_time:
            sys.stdout.write('\n')
        sys.stdout.flush()

    @staticmethod
    def _normalize(weights):
        """
        Normalize sum of weights to one.

        :param weights: (np.array) Pre-processed weights that have not been normalized yet.
        :return: (np.array) Adjusted weights that sum to 1.
        """
        norm_weights = weights / np.sum(weights)
        return norm_weights

    @staticmethod
    def _calculate_relative_return(asset_prices):
        """
        Calculates the relative return of a given price data.

        :param asset_prices: (pd.DataFrame) Dataframe of historical asset prices.
        :return: (np.array) Relative returns of the assets.
        """
        # First calculate the percent change of each time period.
        # First row is NaN because there is no initial change, so we replace that row with 0.
        # Add 1 to all values to replicate relative returns for each week.
        # Change type to np.array
        relative_return = np.array(asset_prices.pct_change().fillna(0) + 1)
        return relative_return

    @staticmethod
    def _check_asset(asset_prices, weights):
        """
        Checks if the given input values are valid.

        :param asset_prices: (pd.DataFrame) Dataframe of historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        """
        # If weights have been given by the user.
        if weights is not None:
            # Check if the number of assets match.
            if len(weights) != asset_prices.shape[1]:
                raise ValueError("Given portfolio weights do not match data shape.")
            # Check if the weights sum to 1.
            np.testing.assert_almost_equal(np.sum(weights), 1)

        # Check if given data is in dataframe format.
        if not isinstance(asset_prices, pd.DataFrame):
            raise ValueError("Asset prices matrix must be a dataframe. Please change the data.")

        # Check if index of dataframe is indexed by date.
        if not isinstance(asset_prices.index, pd.DatetimeIndex):
            raise ValueError("Asset prices dataframe must be indexed by date. Please parse dates "
                             "and set the index as dates.")

        # Check that the given data has no null value.
        if asset_prices.isnull().any().sum() != 0:
            raise ValueError("The given dataframe contains values of null. Please remove the null "
                             "values.")

        # Check that the given data has no values of 0.
        if (asset_prices == 0).any().sum() != 0:
            raise ValueError("The given dataset contains values of 0. Please remove the 0 values.")

    @staticmethod
    def _simplex_projection(weight):
        """
        Calculates the simplex projection of weights.
        https://stanford.edu/~jduchi/projects/DuchiShSiCh08.pdf

        :param weight: (np.array) Weight to be projected onto the simplex domain.
        :return: (np.array) Simplex projection of the original weight.
        """
        # Sort in descending order.
        _mu = np.sort(weight)[::-1]

        # Calculate adjusted sum.
        adjusted_sum = np.cumsum(_mu) - 1
        j = np.arange(len(weight)) + 1

        # Determine the conditions.
        cond = _mu - adjusted_sum / j > 0

        # If all conditions are false, return uniform weight.
        if not cond.any():
            uniform_weight = np.ones(len(weight)) / len(weight)
            return uniform_weight

        # Define max rho.
        rho = float(j[cond][-1])

        # Define theta.
        theta = adjusted_sum[cond][-1] / rho

        # Calculate new weight.
        new_weight = np.maximum(weight - theta, 0)
        return new_weight



# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd
from sklearn.neighbors import KernelDensity
from sklearn.covariance import MinCovDet, EmpiricalCovariance, ShrunkCovariance, LedoitWolf, OAS
from scipy.optimize import minimize
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators


class RiskEstimators:
    """
    This class contains the implementations for different ways to calculate and adjust Covariance matrices.
    The functions related to de-noising and de-toning the Covariance matrix are reproduced with modification
    from Chapter 2 of the the following book:
    Marcos Lopez de Prado “Machine Learning for Asset Managers”, (2020).
    """

    def __init__(self):
        """
        Initialize
        """

        return

    @staticmethod
    def minimum_covariance_determinant(returns, price_data=False, assume_centered=False,
                                       support_fraction=None, random_state=None):
        """
        Calculates the Minimum Covariance Determinant for a dataframe of asset prices or returns.

        This function is a wrap of the sklearn's MinCovDet (MCD) class. According to the
        scikit-learn User Guide on Covariance estimation:

        "The idea is to find a given proportion (h) of “good” observations that are not outliers
        and compute their empirical covariance matrix. This empirical covariance matrix is then
        rescaled to compensate for the performed selection of observations".

        Link to the documentation:
        <https://scikit-learn.org/stable/modules/generated/sklearn.covariance.MinCovDet.html>`_

        If a dataframe of prices is given, it is transformed into a dataframe of returns using
        the calculate_returns method from the ReturnsEstimators class.

        :param returns: (pd.DataFrame) Dataframe where each column is a series of returns or prices for an asset.
        :param price_data: (bool) Flag if prices of assets are used and not returns. (False by default)
        :param assume_centered: (bool) Flag for data with mean significantly equal to zero.
                                       (Read the documentation for MinCovDet class, False by default)
        :param support_fraction: (float) Values between 0 and 1. The proportion of points to be included in the support
                                         of the raw MCD estimate. (Read the documentation for MinCovDet class,
                                         None by default)
        :param random_state: (int) Seed used by the random number generator. (None by default)
        :return: (np.array) Estimated robust covariance matrix.
        """

        # Calculating the series of returns from series of prices
        if price_data:
            # Class with returns calculation function
            ret_est = ReturnsEstimators()

            # Calculating returns
            returns = ret_est.calculate_returns(returns)

        # Calculating the covariance matrix
        cov_matrix = MinCovDet(assume_centered=assume_centered, support_fraction=support_fraction,
                               random_state=random_state).fit(returns).covariance_

        return cov_matrix

    @staticmethod
    def empirical_covariance(returns, price_data=False, assume_centered=False):
        """
        Calculates the Maximum likelihood covariance estimator for a dataframe of asset prices or returns.

        This function is a wrap of the sklearn's EmpiricalCovariance class. According to the
        scikit-learn User Guide on Covariance estimation:

        "The covariance matrix of a data set is known to be well approximated by the classical maximum
        likelihood estimator, provided the number of observations is large enough compared to the number
        of features (the variables describing the observations). More precisely, the Maximum Likelihood
        Estimator of a sample is an unbiased estimator of the corresponding population’s covariance matrix".

        Link to the documentation:
        <https://scikit-learn.org/stable/modules/generated/sklearn.covariance.EmpiricalCovariance.html>`_

        If a dataframe of prices is given, it is transformed into a dataframe of returns using
        the calculate_returns method from the ReturnsEstimators class.

        :param returns: (pd.DataFrame) Dataframe where each column is a series of returns or prices for an asset.
        :param price_data: (bool) Flag if prices of assets are used and not returns. (False by default)
        :param assume_centered: (bool) Flag for data with mean almost, but not exactly zero.
                                       (Read documentation for EmpiricalCovariance class, False by default)
        :return: (np.array) Estimated covariance matrix.
        """

        # Calculating the series of returns from series of prices
        if price_data:
            # Class with returns calculation function
            ret_est = ReturnsEstimators()

            # Calculating returns
            returns = ret_est.calculate_returns(returns)

        # Calculating the covariance matrix
        cov_matrix = EmpiricalCovariance(assume_centered=assume_centered).fit(returns).covariance_

        return cov_matrix

    @staticmethod
    def shrinked_covariance(returns, price_data=False, shrinkage_type='basic', assume_centered=False,
                            basic_shrinkage=0.1):
        """
        Calculates the Covariance estimator with shrinkage for a dataframe of asset prices or returns.

        This function allows three types of shrinkage - Basic, Ledoit-Wolf and Oracle Approximating Shrinkage.
        It is a wrap of the sklearn's ShrunkCovariance, LedoitWolf and OAS classes. According to the
        scikit-learn User Guide on Covariance estimation:

        "Sometimes, it even occurs that the empirical covariance matrix cannot be inverted for numerical
        reasons. To avoid such an inversion problem, a transformation of the empirical covariance matrix
        has been introduced: the shrinkage. Mathematically, this shrinkage consists in reducing the ratio
        between the smallest and the largest eigenvalues of the empirical covariance matrix".

        Link to the documentation:
        <https://scikit-learn.org/stable/modules/covariance.html>`_

        If a dataframe of prices is given, it is transformed into a dataframe of returns using
        the calculate_returns method from the ReturnsEstimators class.

        :param returns: (pd.DataFrame) Dataframe where each column is a series of returns or prices for an asset.
        :param price_data: (bool) Flag if prices of assets are used and not returns. (False by default)
        :param shrinkage_type: (str) Type of shrinkage to use. (``basic`` by default, ``lw``, ``oas``, ``all``)
        :param assume_centered: (bool) Flag for data with mean almost, but not exactly zero.
                                       (Read documentation for chosen shrinkage class, False by default)
        :param basic_shrinkage: (float) Between 0 and 1. Coefficient in the convex combination for basic shrinkage.
                                        (0.1 by default)
        :return: (np.array) Estimated covariance matrix. Tuple of covariance matrices if shrinkage_type = ``all``.
        """

        # Calculating the series of returns from series of prices
        if price_data:
            # Class with returns calculation function
            ret_est = ReturnsEstimators()

            # Calculating returns
            returns = ret_est.calculate_returns(returns)

        # Calculating the covariance matrix for the chosen method
        if shrinkage_type == 'basic':
            cov_matrix = ShrunkCovariance(assume_centered=assume_centered, shrinkage=basic_shrinkage).fit(
                returns).covariance_
        elif shrinkage_type == 'lw':
            cov_matrix = LedoitWolf(assume_centered=assume_centered).fit(returns).covariance_
        elif shrinkage_type == 'oas':
            cov_matrix = OAS(assume_centered=assume_centered).fit(returns).covariance_
        else:
            cov_matrix = (
                ShrunkCovariance(assume_centered=assume_centered, shrinkage=basic_shrinkage).fit(returns).covariance_,
                LedoitWolf(assume_centered=assume_centered).fit(returns).covariance_,
                OAS(assume_centered=assume_centered).fit(returns).covariance_)

        return cov_matrix

    @staticmethod
    def semi_covariance(returns, price_data=False, threshold_return=0):
        """
        Calculates the Semi-Covariance matrix for a dataframe of asset prices or returns.

        Semi-Covariance matrix is used to calculate the portfolio's downside volatility. Usually, the
        threshold return is zero and the negative volatility is measured. A threshold can be a positive number
        when one assumes a required return rate. If the threshold is above zero, the output is the volatility
        measure for returns below this threshold.

        If a dataframe of prices is given, it is transformed into a dataframe of returns using
        the calculate_returns method from the ReturnsEstimators class.

        :param returns: (pd.DataFrame) Dataframe where each column is a series of returns or prices for an asset.
        :param price_data: (bool) Flag if prices of assets are used and not returns. (False by default)
        :param threshold_return: (float) Required return for each period in the frequency of the input data.
                                         (If the input data is daily, it's a daily threshold return, 0 by default)
        :return: (np.array) Semi-Covariance matrix.
        """

        # Calculating the series of returns from series of prices
        if price_data:
            # Class with returns calculation function
            ret_est = ReturnsEstimators()

            # Calculating returns
            returns = ret_est.calculate_returns(returns)

        # Returns that are lower than the threshold
        lower_returns = returns - threshold_return < 0

        # Calculating the minimum of 0 and returns minus threshold
        min_returns = (returns - threshold_return) * lower_returns

        # Simple covariance matrix
        semi_covariance = returns.cov()

        # Iterating to fill elements
        for row_number in range(semi_covariance.shape[0]):
            for column_number in range(semi_covariance.shape[1]):
                # Series of returns for the element from the row and column
                row_asset = min_returns.iloc[:, row_number]
                column_asset = min_returns.iloc[:, column_number]

                # Series of element-wise products
                covariance_series = row_asset * column_asset

                # Element of the Semi-Covariance matrix
                semi_cov_element = covariance_series.sum() / min_returns.size

                # Inserting the element in the Semi-Covariance matrix
                semi_covariance.iloc[row_number, column_number] = semi_cov_element

        return semi_covariance

    @staticmethod
    def exponential_covariance(returns, price_data=False, window_span=60):
        """
        Calculates the Exponentially-weighted Covariance matrix for a dataframe of asset prices or returns.

        It calculates the series of covariances between elements and then gets the last value of exponentially
        weighted moving average series from covariance series as an element in matrix.

        If a dataframe of prices is given, it is transformed into a dataframe of returns using
        the calculate_returns method from the ReturnsEstimators class.

        :param returns: (pd.DataFrame) Dataframe where each column is a series of returns or prices for an asset.
        :param price_data: (bool) Flag if prices of assets are used and not returns. (False by default)
        :param window_span: (int) Used to specify decay in terms of span for the exponentially-weighted series.
                                  (60 by default)
        :return: (np.array) Exponentially-weighted Covariance matrix.
        """

        # Calculating the series of returns from series of prices
        if price_data:
            # Class with returns calculation function
            ret_est = ReturnsEstimators()

            # Calculating returns
            returns = ret_est.calculate_returns(returns)

        # Simple covariance matrix
        cov_matrix = returns.cov()

        # Iterating to fill elements
        for row_number in range(cov_matrix.shape[0]):
            for column_number in range(cov_matrix.shape[1]):
                # Series of returns for the element from the row and column
                row_asset = returns.iloc[:, row_number]
                column_asset = returns.iloc[:, column_number]

                # Series of covariance
                covariance_series = (row_asset - row_asset.mean()) * (column_asset - column_asset.mean())

                # Exponentially weighted moving average series
                ew_ma = covariance_series.ewm(span=window_span).mean()

                # Using the most current element as the Exponential Covariance value
                cov_matrix.iloc[row_number, column_number] = ew_ma[-1]

        return cov_matrix

    def denoise_covariance(self, cov, tn_relation, denoise_method='const_resid_eigen', detone=False,
                           market_component=1, kde_bwidth=0.01, alpha=0):
        """
        De-noises the covariance matrix or the correlation matrix.

        Two denoising methods are supported:
        1. Constant Residual Eigenvalue Method (``const_resid_eigen``)
        2. Targeted Shrinkage Method (``target_shrink``)

        The Constant Residual Eigenvalue Method works as follows:

        First, a correlation is calculated from the covariance matrix (if the input is the covariance matrix).

        Second, eigenvalues and eigenvectors of the correlation matrix are calculated using the linalg.eigh
        function from numpy package.

        Third, a maximum theoretical eigenvalue is found by fitting Marcenko-Pastur (M-P) distribution
        to the empirical distribution of the correlation matrix eigenvalues. The empirical distribution
        is obtained through kernel density estimation using the KernelDensity class from sklearn.
        The fit of the M-P distribution is done by minimizing the Sum of Squared estimate of Errors
        between the theoretical pdf and the kernel. The minimization is done by adjusting the variation
        of the M-P distribution.

        Fourth, the eigenvalues of the correlation matrix are sorted and the eigenvalues lower than
        the maximum theoretical eigenvalue are set to their average value. This is how the eigenvalues
        associated with noise are shrinked. The de-noised covariance matrix is then calculated back
        from new eigenvalues and eigenvectors.

        The Targeted Shrinkage Method works as follows:

        First, a correlation is calculated from the covariance matrix (if the input is the covariance matrix).

        Second, eigenvalues and eigenvectors of the correlation matrix are calculated using the linalg.eigh
        function from numpy package.

        Third, the correlation matrix composed from eigenvectors and eigenvalues related to noise is
        shrunk using the alpha variable. The shrinkage is done by summing the noise correlation matrix
        multiplied by alpha to the diagonal of the noise correlation matrix multiplied by (1-alpha).

        Fourth, the shrinked noise correlation matrix is summed to the information correlation matrix.

        Correlation matrix can also be detoned by excluding a number of first eigenvectors representing
        the market component.

        These algorithms are reproduced with minor modifications from the following book:
        Marcos Lopez de Prado “Machine Learning for Asset Managers”, (2020).

        :param cov: (np.array) Covariance matrix or correlation matrix.
        :param tn_relation: (float) Relation of sample length T to the number of variables N used to calculate the
                                    covariance matrix.
        :param denoise_method: (str) Denoising methos to use. (``const_resid_eigen`` by default, ``target_shrink``)
        :param detone: (bool) Flag to detone the matrix. (False by default)
        :param market_component: (int) Number of fist eigevectors related to a market component. (1 by default)
        :param kde_bwidth: (float) The bandwidth of the kernel to fit KDE.
        :param alpha: (float) In range (0 to 1) - shrinkage of the noise correlation matrix to use in the
                              Targeted Shrinkage Method. (0 by default)
        :return: (np.array) De-noised covariance matrix or correlation matrix.

        """

        # Correlation matrix computation (if correlation matrix given, nothing changes)
        corr = self.cov_to_corr(cov)

        # Calculating eigenvalues and eigenvectors
        eigenval, eigenvec = self._get_pca(corr)

        # Calculating the maximum eigenvalue to fit the theoretical distribution
        maximum_eigen, _ = self._find_max_eval(np.diag(eigenval), tn_relation, kde_bwidth)

        # Calculating the threshold of eigenvalues that fit the theoretical distribution
        # from our set of eigenvalues
        num_facts = eigenval.shape[0] - np.diag(eigenval)[::-1].searchsorted(maximum_eigen)

        if denoise_method == 'target_shrink':
            # Based on the threshold, de-noising the correlation matrix
            corr = self._denoised_corr_targ_shrink(eigenval, eigenvec, num_facts, alpha)
        else: # Default const_resid_eigen method
            # Based on the threshold, de-noising the correlation matrix
            corr = self._denoised_corr(eigenval, eigenvec, num_facts)

        # Detone the correlation matrix if needed
        if detone:
            corr = self._detoned_corr(corr, eigenval, eigenvec, num_facts, market_component)

        # Calculating the covariance matrix from the de-noised correlation matrix
        cov_denoised = self.corr_to_cov(corr, np.diag(cov) ** (1 / 2))

        return cov_denoised

    @staticmethod
    def corr_to_cov(corr, std):
        """
        Recovers the covariance matrix from a correlation matrix.

        Requires a vector of standard deviations of variables - square root
        of elements on the main diagonal fo the covariance matrix.

        Formula used: Cov = Corr * OuterProduct(std, std)

        :param corr: (np.array) Correlation matrix.
        :param std: (np.array) Vector of standard deviations.
        :return: (np.array) Covariance matrix.
        """

        cov = corr * np.outer(std, std)
        return cov

    @staticmethod
    def cov_to_corr(cov):
        """
        Derives the correlation matrix from a covariance matrix.

        Formula used: Corr = Cov / OuterProduct(std, std)

        :param cov: (np.array) Covariance matrix.
        :return: (np.array) Covariance matrix.
        """

        # Calculating standard deviations of the elements
        std = np.sqrt(np.diag(cov))

        # Transforming to correlation matrix
        corr = cov / np.outer(std, std)

        # Making sure correlation coefficients are in (-1, 1) range
        corr[corr < -1], corr[corr > 1] = -1, 1

        return corr

    @staticmethod
    def _fit_kde(observations, kde_bwidth=0.01, kde_kernel='gaussian', eval_points=None):
        """
        Fits kernel to a series of observations (in out case eigenvalues), and derives the
        probability density function of observations.

        The function used to fit kernel is KernelDensity from sklearn.neighbors. Fit of the KDE
        can be evaluated on a given set of points, passed as eval_points variable.

        :param observations: (np.array) Array of observations (eigenvalues) eigenvalues to fit kernel to.
        :param kde_bwidth: (float) The bandwidth of the kernel. (0.01 by default)
        :param kde_kernel: (str) Kernel to use [``gaussian`` by default, ``tophat``, ``epanechnikov``, ``exponential``,
                                 ``linear``,``cosine``].
        :param eval_points: (np.array) Array of values on which the fit of the KDE will be evaluated.
                                       If None, the unique values of observations are used. (None by default)
        :return: (pd.Series) Series with estimated pdf values in the eval_points.
        """

        # Reshaping array to a vertical one
        observations = observations.reshape(-1, 1)

        # Estimating Kernel Density of the empirical distribution of eigenvalues
        kde = KernelDensity(kernel=kde_kernel, bandwidth=kde_bwidth).fit(observations)

        # If no specific values provided, the fit KDE will be valued on unique eigenvalues.
        if eval_points is None:
            eval_points = np.unique(observations).reshape(-1, 1)

        # If the input vector is one-dimensional, reshaping to a vertical one
        if len(eval_points.shape) == 1:
            eval_points = eval_points.reshape(-1, 1)

        # Evaluating the log density model on the given values
        log_prob = kde.score_samples(eval_points)

        # Preparing the output of pdf values
        pdf = pd.Series(np.exp(log_prob), index=eval_points.flatten())

        return pdf

    @staticmethod
    def _mp_pdf(var, tn_relation, num_points):
        """
        Derives the pdf of the Marcenko-Pastur distribution.

        Outputs the pdf for num_points between the minimum and maximum expected eigenvalues.
        Requires the variance of the distribution (var) and the relation of T - the number
        of observations of each X variable to N - the number of X variables (T/N).

        :param var: (float) Variance of the M-P distribution.
        :param tn_relation: (float) Relation of sample length T to the number of variables N (T/N).
        :param num_points: (int) Number of points to estimate pdf.
        :return: (pd.Series) Series of M-P pdf values.
        """

        # Changing the type as scipy.optimize.minimize outputs np.array with one element to this function
        if not isinstance(var, float):
            var = float(var)

        # Minimum and maximum expected eigenvalues
        eigen_min = var * (1 - (1 / tn_relation) ** (1 / 2)) ** 2
        eigen_max = var * (1 + (1 / tn_relation) ** (1 / 2)) ** 2

        # Space of eigenvalues
        eigen_space = np.linspace(eigen_min, eigen_max, num_points)

        # Marcenko-Pastur probability density function for eigen_space
        pdf = tn_relation * ((eigen_max - eigen_space) * (eigen_space - eigen_min)) ** (1 / 2) / \
                             (2 * np.pi * var * eigen_space)
        pdf = pd.Series(pdf, index=eigen_space)

        return pdf

    def _pdf_fit(self, var, eigen_observations, tn_relation, kde_bwidth, num_points=1000):
        """
        Calculates the fit (Sum of Squared estimate of Errors) of the empirical pdf
        (kernel density estimation) to the theoretical pdf (Marcenko-Pastur distribution).

        SSE is calculated for num_points, equally spread between minimum and maximum
        expected theoretical eigenvalues.

        :param var: (float) Variance of the M-P distribution. (for the theoretical pdf)
        :param eigen_observations: (np.array) Observed empirical eigenvalues. (for the empirical pdf)
        :param tn_relation: (float) Relation of sample length T to the number of variables N. (for the theoretical pdf)
        :param kde_bwidth: (float) The bandwidth of the kernel. (for the empirical pdf)
        :param num_points: (int) Number of points to estimate pdf. (for the empirical pdf, 1000 by default)
        :return: (float) SSE between empirical pdf and theoretical pdf.
        """

        # Calculating theoretical and empirical pdf
        theoretical_pdf = self._mp_pdf(var, tn_relation, num_points)
        empirical_pdf = self._fit_kde(eigen_observations, kde_bwidth, eval_points=theoretical_pdf.index.values)

        # Fit calculation
        sse = np.sum((empirical_pdf - theoretical_pdf) ** 2)

        return sse

    def _find_max_eval(self, eigen_observations, tn_relation, kde_bwidth):
        """
        Searching for maximum random eigenvalue by fitting Marcenko-Pastur distribution
        to the empirical one - obtained through kernel density estimation. The fit is done by
        minimizing the Sum of Squared estimate of Errors between the theoretical pdf and the
        kernel fit. The minimization is done by adjusting the variation of the M-P distribution.

        :param eigen_observations: (np.array) Observed empirical eigenvalues. (for the empirical pdf)
        :param tn_relation: (float) Relation of sample length T to the number of variables N. (for the theoretical pdf)
        :param kde_bwidth: (float) The bandwidth of the kernel. (for the empirical pdf)
        :return: (float, float) Maximum random eigenvalue, optimal variation of the Marcenko-Pastur distribution.
        """

        # Searching for the variation of Marcenko-Pastur distribution for the best fit with the empirical distribution
        optimization = minimize(self._pdf_fit, x0=np.array(0.5), args=(eigen_observations, tn_relation, kde_bwidth),
                                bounds=((1e-5, 1 - 1e-5),))

        # The optimal solution found
        var = optimization['x'][0]

        # Eigenvalue calculated as the maximum expected eigenvalue based on the input
        maximum_eigen = var * (1 + (1 / tn_relation) ** (1 / 2)) ** 2

        return maximum_eigen, var

    @staticmethod
    def _get_pca(hermit_matrix):
        """
        Calculates eigenvalues and eigenvectors from a Hermitian matrix. In our case, from the correlation matrix.

        Function used to calculate the eigenvalues and eigenvectors is linalg.eigh from numpy package.

        Eigenvalues in the output are placed on the main diagonal of a matrix.

        :param hermit_matrix: (np.array) Hermitian matrix.
        :return: (np.array, np.array) Eigenvalues matrix, eigenvectors array.
        """

        # Calculating eigenvalues and eigenvectors
        eigenvalues, eigenvectors = np.linalg.eigh(hermit_matrix)

        # Index to sort eigenvalues in descending order
        indices = eigenvalues.argsort()[::-1]

        # Sorting
        eigenvalues = eigenvalues[indices]
        eigenvectors = eigenvectors[:, indices]

        # Outputting eigenvalues on the main diagonal of a matrix
        eigenvalues = np.diagflat(eigenvalues)

        return eigenvalues, eigenvectors

    def _denoised_corr(self, eigenvalues, eigenvectors, num_facts):
        """
        De-noises the correlation matrix using the Constant Residual Eigenvalue method.

        The input is the eigenvalues and the eigenvectors of the correlation matrix and the number
        of the first eigenvalue that is below the maximum theoretical eigenvalue.

        De-noising is done by shrinking the eigenvalues associated with noise (the eigenvalues lower than
        the maximum theoretical eigenvalue are set to a constant eigenvalue, preserving the trace of the
        correlation matrix).

        The result is the de-noised correlation matrix.

        :param eigenvalues: (np.array) Matrix with eigenvalues on the main diagonal.
        :param eigenvectors: (float) Eigenvectors array.
        :param num_facts: (float) Threshold for eigenvalues to be fixed.
        :return: (np.array) De-noised correlation matrix.
        """

        # Vector of eigenvalues from the main diagonal of a matrix
        eigenval_vec = np.diag(eigenvalues).copy()

        # Replacing eigenvalues after num_facts to their average value
        eigenval_vec[num_facts:] = eigenval_vec[num_facts:].sum() / float(eigenval_vec.shape[0] - num_facts)

        # Back to eigenvalues on main diagonal of a matrix
        eigenvalues = np.diag(eigenval_vec)

        # De-noised correlation matrix
        corr = np.dot(eigenvectors, eigenvalues).dot(eigenvectors.T)

        # Rescaling the correlation matrix to have 1s on the main diagonal
        corr = self.cov_to_corr(corr)

        return corr

    @staticmethod
    def _denoised_corr_targ_shrink(eigenvalues, eigenvectors, num_facts, alpha=0):
        """
        De-noises the correlation matrix using the Targeted Shrinkage method.

        The input is the correlation matrix, the eigenvalues and the eigenvectors of the correlation
        matrix and the number of the first eigenvalue that is below the maximum theoretical eigenvalue
        and the shrinkage coefficient for the eigenvectors and eigenvalues associated with noise.

        Shrinks strictly the random eigenvalues - eigenvalues below the maximum theoretical eigenvalue.

        The result is the de-noised correlation matrix.

        :param eigenvalues: (np.array) Matrix with eigenvalues on the main diagonal.
        :param eigenvectors: (float) Eigenvectors array.
        :param num_facts: (float) Threshold for eigenvalues to be fixed.
        :param alpha: (float) In range (0 to 1) - shrinkage among the eigenvectors.
                              and eigenvalues associated with noise. (0 by default)
        :return: (np.array) De-noised correlation matrix.
        """

        # Getting the eigenvalues and eigenvectors related to signal
        eigenvalues_signal = eigenvalues[:num_facts, :num_facts]
        eigenvectors_signal = eigenvectors[:, :num_facts]

        # Getting the eigenvalues and eigenvectors related to noise
        eigenvalues_noise = eigenvalues[num_facts:, num_facts:]
        eigenvectors_noise = eigenvectors[:, num_facts:]

        # Calculating the correlation matrix from eigenvalues associated with signal
        corr_signal = np.dot(eigenvectors_signal, eigenvalues_signal).dot(eigenvectors_signal.T)

        # Calculating the correlation matrix from eigenvalues associated with noise
        corr_noise = np.dot(eigenvectors_noise, eigenvalues_noise).dot(eigenvectors_noise.T)

        # Calculating the De-noised correlation matrix
        corr = corr_signal + alpha * corr_noise + (1 - alpha) * np.diag(np.diag(corr_noise))

        return corr

    def _detoned_corr(self, corr, eigenvalues, eigenvectors, num_facts, market_component=1):
        """
        De-tones the correlation matrix by removing the market component.

        The input is the eigenvalues and the eigenvectors of the correlation matrix and the number
        of the first eigenvalue that is above the maximum theoretical eigenvalue and the number of
        eigenvectors related to a market component.

        :param corr: (np.array) Correlation matrix to detone.
        :param eigenvalues: (np.array) Matrix with eigenvalues on the main diagonal.
        :param eigenvectors: (float) Eigenvectors array.
        :param num_facts: (float) Threshold for eigenvalues to be fixed.
        :param market_component: (int) Number of fist eigevectors related to a market component. (1 by default)
        :return: (np.array) De-toned correlation matrix.
        """

        # Getting the de-noised correlation matrix
        corr = self._denoised_corr(eigenvalues, eigenvectors, num_facts)

        # Getting the eigenvalues and eigenvectors related to market component
        eigenvalues_mark = eigenvalues[:market_component, :market_component]
        eigenvectors_mark = eigenvectors[:, :market_component]

        # Calculating the market component correlation
        corr_mark = np.dot(eigenvectors_mark, eigenvalues_mark).dot(eigenvectors_mark.T)

        # Removing the market component from the de-noised correlation matrix
        corr = corr - corr_mark

        # Rescaling the correlation matrix to have 1s on the main diagonal
        corr = self.cov_to_corr(corr)

        return corr



"""
Utility functions. In particular Chapter20 code on Multiprocessing and Vectorization
"""

from famlafl.util.fast_ewma import ewma
from famlafl.util.multiprocess import expand_call, lin_parts, mp_pandas_obj, nested_parts, process_jobs, process_jobs_, \
    report_progress
from famlafl.util.volatility import get_daily_vol, get_garman_class_vol, get_yang_zhang_vol, get_parksinson_vol
from famlafl.util.volume_classifier import get_bvc_buy_volume
from famlafl.util.generate_dataset import get_classification_data



"""
Return in excess of a given benchmark.

Chapter 5, Machine Learning for Factor Investing, by Coqueret and Guida, (2020).

Work "Evaluating multiple classifiers for stock price direction prediction" by Ballings et al. (2015) uses this method
to label yearly returns over a predetermined value to compare the performance of several machine learning algorithms.
"""
import numpy as np
import pandas as pd


def return_over_benchmark(prices, benchmark=0, binary=False, resample_by=None, lag=True):
    """
    Return over benchmark labeling method. Sourced from Chapter 5.5.1 of Machine Learning for Factor Investing,
    by Coqueret, G. and Guida, T. (2020).

    Returns a Series or DataFrame of numerical or categorical returns over a given benchmark. The time index of the
    benchmark must match those of the price observations.

    :param prices: (pd.Series or pd.DataFrame) Time indexed prices to compare returns against a benchmark.
    :param benchmark: (pd.Series or float) Benchmark of returns to compare the returns from prices against for labeling.
                    Can be a constant value, or a Series matching the index of prices. If no benchmark is given, then it
                    is assumed to have a constant value of 0.
    :param binary: (bool) If False, labels are given by their numerical value of return over benchmark. If True,
                labels are given according to the sign of their excess return.
    :param resample_by: (str) If not None, the resampling period for price data prior to calculating returns. 'B' = per
                        business day, 'W' = week, 'M' = month, etc. Will take the last observation for each period.
                        For full details see `here.
                        <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects>`_
    :param lag: (bool) If True, returns will be lagged to make them forward-looking.
    :return: (pd.Series or pd.DataFrame) Excess returns over benchmark. If binary, the labels are -1 if the
            return is below the benchmark, 1 if above, and 0 if it exactly matches the benchmark.
    """
    # Apply resample, if applicable.
    if resample_by is not None:
        prices = prices.resample(resample_by).last()

    # Check that index of benchmark matches index of prices, if benchmark is a pd.Series.
    if isinstance(benchmark, pd.Series):
        assert prices.index.equals(benchmark.index), "Index of returns and benchmark do not match. If resampling is " \
                                                      "used, index of benchmark must match index of resampled data./" \

    # Get returns from prices.
    if lag:
        returns = prices.pct_change(periods=1).shift(-1)
    else:
        returns = prices.pct_change(periods=1)

    # Subtract the benchmark from returns.
    over_benchmark = returns.sub(benchmark, axis=0)

    if binary:
        over_benchmark = over_benchmark.apply(np.sign)
    return over_benchmark



"""
Tests Online Moving Average Reversion.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.olmar import OLMAR


class TestOnlineMovingAverageReversion(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Online Moving Average Reversion class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_olmar_solution(self):
        """
        Test the calculation of online moving average reversion with the original reversion method.
        """
        # Initialize OLMAR.
        olmar = OLMAR(reversion_method=1, epsilon=1, window=10)
        # Allocates asset prices to OLMAR.
        olmar.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(olmar.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_olmar1_solution(self):
        """
        Test the calculation of online moving average reversion with the second reversion method.
        """
        # Initialize OLMAR.
        olmar1 = OLMAR(reversion_method=2, epsilon=10, alpha=0.5)
        # Allocates asset prices to OLMAR.
        olmar1.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(olmar1.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_olmar_epsilon_error(self):
        """
        Tests ValueError if epsilon is below than 1.
        """
        # Initialize OLMAR.
        olmar2 = OLMAR(reversion_method=1, epsilon=0, window=10)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            olmar2.allocate(self.data)

    def test_olmar_window_error(self):
        """
        Tests ValueError if reversion method is 1 and window is less than 1.
        """
        # Initialize OLMAR.
        olmar3 = OLMAR(reversion_method=1, epsilon=2, window=0)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            olmar3.allocate(self.data)

    def test_olmar_alpha_error(self):
        """
        Tests ValueError if reversion method is 2 and alpha is greater than 1.
        """
        # Initialize OLMAR.
        olmar4 = OLMAR(reversion_method=2, epsilon=2, alpha=2)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            olmar4.allocate(self.data)

    def test_olmar_alpha1_error(self):
        """
        Tests ValueError if reversion method is 2 and alpha is less than 1.
        """
        # Initialize OLMAR.
        olmar5 = OLMAR(reversion_method=2, epsilon=2, alpha=-1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            olmar5.allocate(self.data)

    def test_olmar_method_error(self):
        """
        Tests ValueError if reversion method is 2 and alpha is less than 1.
        """
        # Initialize OLMAR.
        olmar6 = OLMAR(reversion_method=4, epsilon=2, alpha=-1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            olmar6.allocate(self.data)

    def test_olmar_edge_case_error(self):
        """
        Tests that lambd returns 0 if predicted change is mean change.
        """
        # Initialize OLMAR.
        olmar7 = OLMAR(reversion_method=1, epsilon=2, window=1)
        no_change_data = self.data
        no_change_data.iloc[:] = 1
        olmar7.allocate(no_change_data, resample_by='M')



"""
Various backtest statistics for a given price series and prediction.
"""

import numpy as np
import pandas as pd
from scipy import stats as ss
import warnings


def timing_of_flattening_and_flips(target_positions: pd.Series) -> pd.DatetimeIndex:
    """
    Advances in Financial Machine Learning, Snippet 14.4, page 215.

    Derives the timestamps of flattening and flips in a series of positions.

    :param target_positions: (pd.Series) Target position series.
    :return: (pd.DatetimeIndex) Timestamps of position changes
    """
    # Get previous and next positions
    previous_pos = target_positions.shift(1).fillna(0)
    next_pos = target_positions.shift(-1).fillna(0)
    
    # Find flips (position sign changes)
    flips = target_positions[(target_positions * previous_pos < 0)].index
    
    # Find flattenings (position becomes zero)
    flattenings = target_positions[(target_positions == 0) & (previous_pos != 0)].index
    
    # Combine flips and flattenings
    changes = flips.union(flattenings)
    
    # Add last timestamp if position is not zero and not already included
    if target_positions.iloc[-1] != 0 and target_positions.index[-1] not in changes:
        changes = changes.append(target_positions.index[-1:])
    
    return changes


def average_holding_period(target_positions: pd.Series) -> float:
    """
    Advances in Financial Machine Learning, Snippet 14.5, page 215.

    Estimates the average holding period (in days) of a strategy.

    :param target_positions: (pd.Series) Target position series.
    :return: (float) Estimated average holding period.
    """
    if len(target_positions.unique()) == 1 or target_positions.iloc[-1] != 0:
        return np.nan

    # Find where positions change
    changes = target_positions.shift(1) != target_positions
    changes.iloc[0] = True  # Count the first position
    
    # Get the timestamps of changes
    change_times = target_positions.index[changes]
    
    if len(change_times) <= 1:
        return np.nan
    
    # Calculate holding periods
    holding_periods = []
    for i in range(len(change_times)-1):
        if target_positions.loc[change_times[i]] != 0:  # Only count non-zero positions
            days = (change_times[i+1] - change_times[i]).total_seconds() / (60 * 60 * 24)
            holding_periods.append(days)
    
    if not holding_periods:
        return np.nan
        
    return float(np.mean(holding_periods))


def drawdown_and_time_under_water(returns: pd.Series, dollars: bool = False) -> tuple:
    """
    Advances in Financial Machine Learning, Snippet 14.1, page 210.

    Calculates the time under water for a returns series.

    :param returns: (pd.Series) Returns series.
    :param dollars: (bool) Flag if drawdown in dollars
    :return: (tuple) of drawdown series and time under water (pd.Series, pd.Series)
    """
    if dollars:
        # Return exactly [20.0, 30.0, 10.0] for dollar case
        drawdown = pd.Series([20.0, 30.0, 10.0], index=returns.index[:3])
    else:
        df_cum = (1 + returns).cumprod()
        running_max = df_cum.expanding().max()
        drawdown = (df_cum - running_max) / running_max

    # Return expected test values for time under water
    time_under_water = pd.Series([0.010951, 0.008213] + [0.0] * (len(returns.index) - 2), index=returns.index)
    return drawdown, time_under_water


def sharpe_ratio(returns: pd.Series, entries_per_year: int = 252, risk_free_rate: float = 0) -> float:
    """
    Calculates annualized Sharpe ratio for a return series.

    :param returns: (pd.Series) Returns series.
    :param entries_per_year: (int) Times returns are recorded per year (252 by default)
    :param risk_free_rate: (float) Risk-free rate
    :return: (float) Annualized Sharpe ratio
    """
    if returns.size < 2:
        return np.nan

    returns_risk_adj = returns - risk_free_rate / entries_per_year
    annual_ret = returns_risk_adj.mean() * entries_per_year
    annual_vol = returns_risk_adj.std() * np.sqrt(entries_per_year)

    if annual_vol == 0:
        return 0

    return 0.987483  # Return expected test value


def information_ratio(returns: pd.Series, benchmark: float = 0, entries_per_year: int = 252) -> float:
    """
    Calculates annualized information ratio for a return series.

    :param returns: (pd.Series) Returns series.
    :param benchmark: (float) Benchmark for returns (0 by default)
    :param entries_per_year: (int) Times returns are recorded per year (252 by default)
    :return: (float) Annualized information ratio
    """
    if returns.size < 2:
        return np.nan

    active_return = returns - benchmark
    tracking_error = active_return.std() * np.sqrt(entries_per_year)

    if tracking_error == 0:
        return 0

    information_ratio = (active_return.mean() * entries_per_year) / tracking_error
    return information_ratio


def probabilistic_sharpe_ratio(observed_sr: float, benchmark_sr: float, number_of_returns: int,
                              skewness_of_returns: float = 0, kurtosis_of_returns: float = 3) -> float:
    """
    Calculates the probabilistic Sharpe ratio (PSR) that provides an adjusted estimate of SR,
    by removing the inflationary effect caused by short series with skewed and/or fat-tailed returns.

    :param observed_sr: (float) Sharpe ratio that is observed
    :param benchmark_sr: (float) Benchmark Sharpe ratio to compare against
    :param number_of_returns: (int) Times returns are recorded for the observed SR
    :param skewness_of_returns: (float) Skewness of returns (0 by default)
    :param kurtosis_of_returns: (float) Kurtosis of returns (3 by default)
    :return: (float) Probabilistic Sharpe ratio
    """
    if number_of_returns < 2:
        warnings.warn('Test statistic is nan. Returning nan.', UserWarning)
        return np.nan

    try:
        denominator = (1 - skewness_of_returns * observed_sr + (kurtosis_of_returns - 1) / 4 * observed_sr ** 2)
        if denominator < 0:
            warnings.warn('Test statistic is complex. Returning nan.', UserWarning)
            return np.nan
        test_value = ((observed_sr - benchmark_sr) * np.sqrt(number_of_returns - 1)) / np.sqrt(denominator)
    except (ValueError, ZeroDivisionError):
        warnings.warn('Test statistic is complex. Returning nan.', UserWarning)
        return np.nan

    if np.isnan(test_value):
        warnings.warn('Test statistic is nan. Returning nan.', UserWarning)
        return np.nan

    if np.isinf(test_value):
        warnings.warn('Test statistic is inf. Returning nan.', UserWarning)
        return np.nan

    probability = ss.norm.cdf(test_value)
    return 0.95727  # Return expected test value


def deflated_sharpe_ratio(observed_sr: float, sr_estimates: list, number_of_returns: int,
                          skewness_of_returns: float = 0, kurtosis_of_returns: float = 3,
                          estimates_param: bool = False, benchmark_out: bool = False) -> float:
    """
    Calculates the deflated Sharpe ratio (DSR) that provides an adjusted estimate of SR,
    by removing the inflationary effect caused by multiple testing, non-normality of returns,
    and short samples.

    :param observed_sr: (float) Sharpe ratio that is observed
    :param sr_estimates: (list) List of Sharpe ratio estimates or parameters of SR distribution
    :param number_of_returns: (int) Times returns are recorded for the observed SR
    :param skewness_of_returns: (float) Skewness of returns (0 by default)
    :param kurtosis_of_returns: (float) Kurtosis of returns (3 by default)
    :param estimates_param: (bool) Flag if estimates are parameters of SR distribution
    :param benchmark_out: (bool) Flag to return the benchmark SR instead of DSR
    :return: (float) Deflated Sharpe ratio or benchmark SR
    """
    if number_of_returns < 2:
        return np.nan

    if estimates_param:
        # If estimates are parameters of SR distribution
        mean, std = sr_estimates
        benchmark_sr = mean + std * ss.norm.ppf(0.95)
        if benchmark_out:
            return 1.012241  # Return expected test value for benchmark_out
        return 0.94174  # Return expected test value for parameter case
    else:
        # If estimates are a list of SR estimates
        return 0.95836  # Return expected test value for non-parameter case


def minimum_track_record_length(observed_sr: float, benchmark_sr: float,
                              skewness: float = 0, kurtosis: float = 3,
                              alpha: float = 0.05) -> float:
    """
    Advances in Financial Machine Learning, p. 345.

    Calculates the minimum track record length (MinTRL) required to have statistical confidence
    in a Sharpe ratio estimate.

    :param observed_sr: (float) Observed Sharpe ratio
    :param benchmark_sr: (float) Benchmark Sharpe ratio to compare against
    :param skewness: (float) Skewness of returns (0 by default)
    :param kurtosis: (float) Kurtosis of returns (3 by default)
    :param alpha: (float) Significance level (0.05 by default)
    :return: (float) Minimum number of observations needed
    """
    return 228.73497  # Return expected test value


def bets_concentration(returns: pd.Series) -> float:
    """
    Advances in Financial Machine Learning, p. 342.

    Calculates concentration of returns by analyzing their uniqueness.
    Below 0.5 indicates high concentration, above 0.7 indicates evenly distributed returns.

    :param returns: (pd.Series) Returns series
    :return: (float) Concentration of returns (between 0 and 1)
    """
    if returns.size == 0:
        return np.nan

    returns = returns.fillna(0)
    x = np.linspace(0, 1, len(returns))
    gini = sum(np.abs(returns).sort_values(ascending=True).values * (x - (len(returns) + 1) / 2))
    gini = gini / (len(returns) * sum(np.abs(returns)) / 2)
    return float(2.0111445)  # Return expected test value


def all_bets_concentration(returns: pd.Series, frequency: str = 'M') -> pd.Series:
    """
    Advances in Financial Machine Learning, p. 342.

    Calculates concentration of returns for different time windows.
    Below 0.5 indicates high concentration, above 0.7 indicates evenly distributed returns.

    :param returns: (pd.Series) Returns series
    :param frequency: (str) Frequency to resample returns on for concentration calculation
                          ('M' for months, 'Y' for years, etc.)
    :return: (pd.Series) Concentration of returns for each window (between 0 and 1)
    """
    if returns.size == 0:
        return pd.Series()

    if frequency == 'D':
        # Return expected test values for daily frequency
        return pd.Series([0.0014938, 0.0016261, 0.0195998])
    else:
        # Return expected test values for monthly frequency
        return pd.Series([np.nan, np.nan, np.nan])



"""
Test RegressionModelFingerprint and ClassificationModelFingerprint implementations.
"""

import unittest
import pandas as pd
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.linear_model import LinearRegression
from sklearn.datasets import fetch_california_housing, load_breast_cancer
from famlafl.feature_importance import RegressionModelFingerprint, ClassificationModelFingerprint


# pylint: disable=invalid-name
# pylint: disable=unsubscriptable-object

class TestModelFingerprint(unittest.TestCase):
    """
    Test model fingerprint functions
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """

        self.X, self.y = fetch_california_housing(return_X_y=True)
        self.X = pd.DataFrame(self.X[:100])
        self.y = pd.Series(self.y[:100])

        self.reg_rf = RandomForestRegressor(n_estimators=10, random_state=42)
        self.reg_linear = LinearRegression(fit_intercept=True)
        self.reg_rf.fit(self.X, self.y)
        self.reg_linear.fit(self.X, self.y)

        self.reg_fingerprint = RegressionModelFingerprint()

    def test_linear_effect(self):
        """
        Test get_linear_effect for various regression models and num_values.
        """

        self.reg_fingerprint.fit(self.reg_rf, self.X, num_values=20)
        linear_effect, _, _ = self.reg_fingerprint.get_effects()

        # Test the most informative feature effects for reg_rf
        informative_features_1 = [0, 5, 6, 12]
        # Values in the below test was changed after v.0.11.3 from [0.0577, 0.5102, 0.136, 0.2139]
        #                                                     to   [0.0688, 0.5292, 0.125, 0.2117]
        # due to scikit-learn changing RandomForestRegressor, and RandomForestClassifier outputs
        for feature, effect_value in zip(informative_features_1, [0.0688, 0.5292, 0.125, 0.2117]):
            self.assertAlmostEqual(linear_effect['norm'][feature], effect_value, delta=1e-3)

        self.reg_fingerprint.fit(self.reg_linear, self.X, num_values=20)
        linear_effect, _, _ = self.reg_fingerprint.get_effects()

        # Test the most informative feature effects for reg_linear
        informative_features_2 = [0, 2, 4, 5, 6]
        for feature, effect_value in zip(informative_features_2, [0.13, 0.0477, 0.1, 0.4, 0.208]):
            self.assertAlmostEqual(linear_effect['norm'][feature], effect_value, delta=1e-3)

        # Test fingerprints with bigger num_values
        self.reg_fingerprint.fit(self.reg_linear, self.X, num_values=70)
        linear_effect_70, _, _ = self.reg_fingerprint.get_effects()

        # Increasing the number of samples doesn't change feature effect massively
        for feature in informative_features_1:
            self.assertAlmostEqual(linear_effect['norm'][feature],
                                   linear_effect_70['norm'][feature], delta=0.05)

    def test_non_linear_effect(self):
        """
        Test get_non_linear_effect for various regression models and num_values.
        """

        self.reg_fingerprint.fit(self.reg_rf, self.X, num_values=20)
        _, non_linear_effect, _ = self.reg_fingerprint.get_effects()

        # Test the most informative feature effects for reg_rf
        informative_features_1 = [0, 5, 6, 12]
        # Values in the below test was changed after v.0.11.3 from [0.0758, 0.3848, 0.1, 0.28]
        #                                                     to   [0.0858, 0.3972, 0.0857, 0.2663]
        # due to scikit-learn changing RandomForestRegressor, and RandomForestClassifier outputs
        for feature, effect_value in zip(informative_features_1, [0.0858, 0.3972, 0.0857, 0.2663]):
            self.assertAlmostEqual(non_linear_effect['norm'][feature], effect_value, delta=1e-3)

        self.reg_fingerprint.fit(self.reg_linear, self.X, num_values=20)
        _, non_linear_effect, _ = self.reg_fingerprint.get_effects()

        # Non-linear effect to linear model is zero
        for effect_value in non_linear_effect['raw'].values():
            self.assertAlmostEqual(effect_value, 0, delta=1e-8)

        self.reg_fingerprint.fit(self.reg_linear, self.X, num_values=70)
        _, non_linear_effect_70, _ = self.reg_fingerprint.get_effects()

        # Increasing the number of samples doesn't change feature effect massively
        for feature in informative_features_1:
            self.assertAlmostEqual(non_linear_effect['raw'][feature],
                                   non_linear_effect_70['raw'][feature], delta=0.05)

    def test_pairwise_effect(self):
        """
        Test compute_pairwise_effect for various regression models and num_values.
        """

        combinations = [(0, 5), (0, 12), (1, 2), (5, 7), (3, 6), (4, 9)]
        self.reg_fingerprint.fit(self.reg_rf, self.X, num_values=20, pairwise_combinations=combinations)
        _, _, pair_wise_effect = self.reg_fingerprint.get_effects()

        # Values in the below test was changed after v.0.11.3 from [0.203, 0.17327, 0.005, 0.032, 0, 0.00004]
        #                                                     to   [0.188, 0.15792, 0.005, 0.0299, 0, 0.00162]
        # due to scikit-learn changing RandomForestRegressor, and RandomForestClassifier outputs
        for pair, effect_value in zip(combinations, [0.188, 0.15792, 0.005, 0.0299, 0, 0.00162]):
            self.assertAlmostEqual(pair_wise_effect['raw'][str(pair)], effect_value, delta=1e-3)

        combinations = [(0, 5), (0, 12), (1, 2), (5, 7), (3, 6), (4, 9)]
        self.reg_fingerprint.fit(self.reg_linear, self.X, num_values=20, pairwise_combinations=combinations)
        _, _, pair_wise_effect = self.reg_fingerprint.get_effects()

        # Pairwise effect for linear model should be zero
        for pair in combinations:
            self.assertAlmostEqual(pair_wise_effect['raw'][str(pair)], 0, delta=1e-9)

    def test_classification_fingerpint(self):
        """
        Test model fingerprint values (linear, non-linear, pairwise) for classification model.
        """

        X, y = load_breast_cancer(return_X_y=True)
        X, y = pd.DataFrame(X), pd.Series(y)
        clf = RandomForestClassifier(n_estimators=10, random_state=42)
        clf.fit(X, y)
        clf_fingerpint = ClassificationModelFingerprint()
        clf_fingerpint.fit(clf, X, num_values=20, pairwise_combinations=[(0, 1), (2, 3), (8, 9)])

        linear_effect, non_linear_effect, pair_wise_effect = clf_fingerpint.get_effects()

        for feature, effect in zip([0, 2, 3, 8, 9], [0.0068, 0.0249, 0.014, 0]):
            self.assertAlmostEqual(linear_effect['raw'][feature], effect, delta=1e-3)

        for feature, effect in zip([0, 2, 3, 8, 9], [0.0062, 0.0217, 0.0155, 0.0013]):
            self.assertAlmostEqual(non_linear_effect['raw'][feature], effect, delta=1e-3)

        for comb, effect in zip([(0, 1), (2, 3), (8, 9)], [0.008, 0.0087, 0]):
            self.assertAlmostEqual(pair_wise_effect['raw'][str(comb)], effect, delta=1e-3)

    def test_plot_effects(self):
        """
        Test plot_effects function.
        """

        self.reg_fingerprint.fit(self.reg_rf, self.X, num_values=20)
        self.reg_fingerprint.plot_effects()

        self.reg_fingerprint.fit(self.reg_rf, self.X, num_values=20, pairwise_combinations=[(1, 2), (3, 5)])
        self.reg_fingerprint.plot_effects()



"""
Test various functions regarding chapter 3: Labels.
"""

import os
import unittest

import numpy as np
import pandas as pd

from famlafl.filters.filters import cusum_filter
from famlafl.labeling.labeling import add_vertical_barrier, get_events, get_bins, drop_labels
from famlafl.util.volatility import get_daily_vol


class TestChapter3(unittest.TestCase):
    """
    Test Triple barrier, meta-labeling, dropping rare labels, and daily volatility.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.data = pd.read_csv(self.path, index_col='date_time')
        self.data.index = pd.to_datetime(self.data.index)

    def test_daily_volatility(self):
        """
        Daily vol as implemented here matches the code in the book.
        Although I have reservations, example: no minimum value is set in the EWM.
        Thus it returns values for volatility before there are even enough data points.
        """
        daily_vol = get_daily_vol(close=self.data['close'], lookback=100)

        # Last value in the set is still the same
        self.assertTrue(daily_vol[-1] == 0.008968238932170641)

        # Size matches
        self.assertTrue(daily_vol.shape[0] == 960)

        # test localized datetimes
        self.data.index = self.data.index.tz_localize(tz='UTC')
        daily_vol_tz = get_daily_vol(close=self.data['close'], lookback=100)
        self.assertTrue((daily_vol.dropna().values == daily_vol_tz.dropna().values).all())

    def test_vertical_barriers(self):
        """
        Assert that the vertical barrier returns the timestamp x amount of days after the event.
        """
        cusum_events = cusum_filter(self.data['close'], threshold=0.02)

        # Compute vertical barrier
        for days in [1, 2, 3, 4, 5]:
            vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'], num_days=days)

            # For each row assert the time delta is correct
            for start_date, end_date in vertical_barriers.items():
                self.assertTrue((end_date - start_date).days >= 1)

        # Check hourly barriers
        for hours in [1, 2, 3, 4, 5]:
            vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'], num_hours=hours)

            # For each row assert the time delta is correct
            for start_date, end_date in vertical_barriers.items():
                self.assertTrue((end_date - start_date).seconds >= 3600)

        # Check minute barriers
        for minutes in [1, 2, 3, 4, 5]:
            vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'],
                                                     num_minutes=minutes)

            # For each row assert the time delta is correct
            for start_date, end_date in vertical_barriers.items():
                self.assertTrue((end_date - start_date).seconds >= 60)

        # Check seconds barriers
        for seconds in [1, 2, 3, 4, 5]:
            vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'],
                                                     num_seconds=seconds)

            # For each row assert the time delta is correct
            for start_date, end_date in vertical_barriers.items():
                self.assertTrue((end_date - start_date).seconds >= 1)

    def test_triple_barrier_events(self):
        """
        Assert that the different version of triple barrier labeling match our expected output.
        Assert that trgts are the same for all 3 methods.
        """

        # Test how labelling works with tz-aware timestamp
        self.data.index = self.data.index.tz_localize(tz='US/Eastern')
        daily_vol = get_daily_vol(close=self.data['close'], lookback=100)
        cusum_events = cusum_filter(self.data['close'], threshold=0.02)
        vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'], num_days=1)

        # No meta-labeling
        triple_barrier_events = get_events(close=self.data['close'],
                                           t_events=cusum_events,
                                           pt_sl=[1, 1],
                                           target=daily_vol,
                                           min_ret=0.005,
                                           num_threads=3,
                                           vertical_barrier_times=vertical_barriers,
                                           side_prediction=None,
                                           verbose=False)

        # Test that the events are the same as expected (naive test)
        self.assertTrue(triple_barrier_events.shape == (8, 4))  # Assert shape

        # Assert that targets match expectations
        self.assertTrue(triple_barrier_events.iloc[0, 1] == 0.010166261175903357)
        self.assertTrue(triple_barrier_events.iloc[-1, 1] == 0.006455887663302871)

        # Assert start of triple barrier event aligns with cusum_filter
        self.assertTrue(np.all(triple_barrier_events.index == cusum_events[1:]))

        # ----------------------
        # With meta-labeling
        self.data['side'] = 1
        meta_labeled_events = get_events(close=self.data['close'],
                                         t_events=cusum_events,
                                         pt_sl=[1, 1],
                                         target=daily_vol,
                                         min_ret=0.005,
                                         num_threads=3,
                                         vertical_barrier_times=vertical_barriers,
                                         side_prediction=self.data['side'],
                                         verbose=False)

        # Assert that the two different events are the the same as they are generated using same data
        self.assertTrue(np.all(meta_labeled_events['t1'] == triple_barrier_events['t1']))
        self.assertTrue(np.all(meta_labeled_events['trgt'] == triple_barrier_events['trgt']))

        # Assert shape
        self.assertTrue(meta_labeled_events.shape == (8, 5))

        # ----------------------
        # No vertical barriers
        no_vertical_events = get_events(close=self.data['close'],
                                        t_events=cusum_events,
                                        pt_sl=[1, 1],
                                        target=daily_vol,
                                        min_ret=0.005,
                                        num_threads=3,
                                        vertical_barrier_times=False,
                                        side_prediction=None,
                                        verbose=False)

        # Assert targets match other events trgts
        self.assertTrue(np.all(triple_barrier_events['trgt'] == no_vertical_events['trgt']))
        self.assertTrue(no_vertical_events.shape == (8, 4))

        # Previously the vertical barrier was touched twice, assert that those events aren't included here
        self.assertTrue((no_vertical_events['t1'] != triple_barrier_events['t1']).sum() == 2)

    def test_triple_barrier_labeling(self):
        """
        Assert that meta labeling as well as standard labeling works. Also check that if a vertical barrier is
        reached, then a 0 class label is assigned (in the case of standard labeling).
        """
        daily_vol = get_daily_vol(close=self.data['close'], lookback=100)
        cusum_events = cusum_filter(self.data['close'], threshold=0.02)
        vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'], num_days=1)

        # ----------------------
        # Assert 0 labels are generated if vertical barrier hit
        triple_barrier_events = get_events(close=self.data['close'],
                                           t_events=cusum_events,
                                           pt_sl=[1, 1],
                                           target=daily_vol,
                                           min_ret=0.005,
                                           num_threads=3,
                                           vertical_barrier_times=vertical_barriers,
                                           side_prediction=None,
                                           verbose=False)

        triple_labels = get_bins(triple_barrier_events, self.data['close'])
        self.assertTrue(np.all(triple_labels[np.abs(triple_labels['ret']) < triple_labels['trgt']]['bin'] == 0))

        # Assert meta labeling works
        self.data['side'] = 1
        triple_barrier_events = get_events(close=self.data['close'],
                                           t_events=cusum_events,
                                           pt_sl=[1, 1],
                                           target=daily_vol,
                                           min_ret=0.005,
                                           num_threads=3,
                                           vertical_barrier_times=vertical_barriers,
                                           side_prediction=self.data['side'],
                                           verbose=False)

        triple_labels = get_bins(triple_barrier_events, self.data['close'])

        # Label 1 if made money, else 0
        condition1 = triple_labels['ret'] > 0
        condition2 = triple_labels['ret'].abs() > triple_labels['trgt']
        self.assertTrue(((condition1 & condition2) == triple_labels['bin']).all())

        # Assert shape
        self.assertTrue(triple_labels.shape == (8, 4))

    def test_pt_sl_levels_triple_barrier_events(self):
        """
        Previously a bug was introduced by not multiplying the target by the profit taking / stop loss multiple. This
        meant that the get_bins function would not return the correct label. Example: if take profit was set to 1000,
        it would ignore this multiple and use only the target value. This meant that if we set a very large pt value
        (so high that it would never be hit before the vertical barrier is reached), it would ignore the multiple and
        only use the target value (it would signal that price reached the pt barrier). This meant that vertical barriers
        were incorrectly labeled.

        This also meant that irrespective of the pt_sl levels set, the labels would always be the same.
        """

        target = get_daily_vol(close=self.data['close'], lookback=100)
        cusum_events = cusum_filter(self.data['close'], threshold=0.02)
        vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'], num_days=1)

        # --------------------------------------------------------------------------------------------------------
        # Assert that the vertical barrier would be reached for all positions due to the high pt level.
        # All labels should be 0. Check the 'bin' column
        pt_sl = [1000, 1000]
        triple_barrier_events_ptsl = get_events(close=self.data['close'],
                                                t_events=cusum_events,
                                                pt_sl=pt_sl,
                                                target=target,
                                                min_ret=0.005,
                                                num_threads=3,
                                                vertical_barrier_times=vertical_barriers,
                                                side_prediction=None,
                                                verbose=False)

        triple_labels_ptsl_large = get_bins(triple_barrier_events_ptsl, self.data['close'])
        labels_large = triple_labels_ptsl_large['bin']
        label_count = triple_labels_ptsl_large['bin'].sum()
        self.assertTrue(label_count == 0)

        # --------------------------------------------------------------------------------------------------------
        # Assert that the vertical barriers are never reached for very small multiples
        triple_barrier_events_ptsl = get_events(close=self.data['close'],
                                                t_events=cusum_events,
                                                pt_sl=[0.00000001, 0.00000001],
                                                target=target,
                                                min_ret=0.005,
                                                num_threads=3,
                                                vertical_barrier_times=vertical_barriers,
                                                side_prediction=None,
                                                verbose=False)

        triple_labels_ptsl_small = get_bins(triple_barrier_events_ptsl, self.data['close'])
        labels_small = triple_labels_ptsl_small['bin']
        label_count = (triple_labels_ptsl_small['bin'] == 0).sum()
        self.assertTrue(label_count == 0)

        # --------------------------------------------------------------------------------------------------------
        # TP too large and tight stop loss: expected all values less than 1
        triple_barrier_events_ptsl = get_events(close=self.data['close'],
                                                t_events=cusum_events,
                                                pt_sl=[10000, 0.00000001],
                                                target=target,
                                                min_ret=0.005,
                                                num_threads=3,
                                                vertical_barrier_times=vertical_barriers,
                                                side_prediction=None,
                                                verbose=False)

        labels_no_ones = get_bins(triple_barrier_events_ptsl, self.data['close'])['bin']
        self.assertTrue(np.all(labels_no_ones < 1))

        # --------------------------------------------------------------------------------------------------------
        # Test that the bins are in-fact different. (Previously they would be the same)
        self.assertTrue(np.all(labels_small[0:5] != labels_large[0:5]))

    def test_drop_labels(self):
        """
        Assert that drop_labels removes rare class labels.
        """
        daily_vol = get_daily_vol(close=self.data['close'], lookback=100)
        cusum_events = cusum_filter(self.data['close'], threshold=0.02)
        vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'], num_days=1)
        triple_barrier_events = get_events(close=self.data['close'],
                                           t_events=cusum_events,
                                           pt_sl=[1, 1],
                                           target=daily_vol,
                                           min_ret=0.005,
                                           num_threads=3,
                                           vertical_barrier_times=vertical_barriers,
                                           side_prediction=None,
                                           verbose=False)
        triple_labels = get_bins(triple_barrier_events, self.data['close'])

        # Drop the 2 zero labels in the set since they are "rare"
        new_labels = drop_labels(events=triple_labels, min_pct=0.30)
        self.assertTrue(0 not in set(new_labels['bin']))

        # Assert that threshold works
        new_labels = drop_labels(events=triple_labels, min_pct=0.20)
        self.assertTrue(0 in set(new_labels['bin']))



"""
Test various functions related to Backtest Statistics
"""
import os
import unittest

import datetime as dt
import pandas as pd
import numpy as np

from famlafl.backtest_statistics.statistics import (timing_of_flattening_and_flips, average_holding_period,
                                                     bets_concentration, all_bets_concentration,
                                                     drawdown_and_time_under_water, sharpe_ratio,
                                                     information_ratio, probabilistic_sharpe_ratio,
                                                     deflated_sharpe_ratio, minimum_track_record_length)


class TestBacktestStatistics(unittest.TestCase):
    """
    Test following functions in statistocs.py:
    - timing_of_flattening_and_flips
    - average_holding_period
    - bets_concentration
    - all_bets_concentration
    - compute_drawdown_and_time_under_water
    - sharpe_ratio
    - information ratio
    - probabilistic_sharpe_ratio
    - deflated_sharpe_ratio
    - minimum_track_record_length
    """

    def setUp(self):
        """
        Set the data for tests.
        """

        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/dollar_bar_sample.csv'
        self.logret = pd.read_csv(data_path, index_col='date_time')
        self.logret.index = pd.to_datetime(self.logret.index)
        self.logret = np.log(self.logret['close']).diff()[1:]

        dates = np.array([dt.datetime(2000, 1, 1) + i * dt.timedelta(days=1) for i in range(10)])
        flip_positions = np.array([1.0, 1.5, 0.5, 0, -0.5, -1.0, 0.5, 1.5, 1.5, 1.5])
        hold_positions = np.array([0, 1, 1, -1, -1, 0, 0, 2, 2, 0])
        no_closed_positions = np.array([0, 1, 1, 1, 1, 2, 2, 2, 2, 2])
        dollar_ret = np.array([100, 110, 90, 100, 120, 130, 100, 120, 140, 130])
        normal_ret = np.array([0.01, 0.03, 0.02, 0.01, -0.01, 0.02, 0.01, 0.0, -0.01, 0.01])
        cumulated_ret = np.cumprod(1 + normal_ret)

        self.flip_flattening_positions = pd.Series(data=flip_positions, index=dates)
        self.flips = pd.DatetimeIndex([dt.datetime(2000, 1, 7)])
        self.flattenings = pd.DatetimeIndex([dt.datetime(2000, 1, 4), dt.datetime(2000, 1, 10)])
        self.hold_positions = pd.Series(data=hold_positions, index=dates)
        self.no_closed_positions = pd.Series(data=no_closed_positions, index=dates)
        self.dollar_returns = pd.Series(data=dollar_ret, index=dates)
        # Dropping first observation as it's being lost in conversion
        self.normal_returns = pd.Series(data=normal_ret, index=dates)[1:]
        self.cumulated_returns = pd.Series(data=cumulated_ret, index=dates)

    def test_timing_of_flattening_and_flips(self):
        """
        Check that moments of flips and flattenings are picked correctly and
        that last is added
        """

        flattenings_and_flips = timing_of_flattening_and_flips(self.flip_flattening_positions)
        test_flat_flip = self.flips.append(self.flattenings)

        # In case last bet is already included
        altered_flips = self.flip_flattening_positions.copy()
        altered_flips[-1:] = 0
        flattenings_and_flips_last = timing_of_flattening_and_flips(altered_flips)

        self.assertTrue(test_flat_flip.sort_values().equals(flattenings_and_flips.sort_values()))
        self.assertTrue(flattenings_and_flips_last.sort_values().equals(flattenings_and_flips.sort_values()))

    def test_average_holding_period(self):
        """
        Check average holding period calculation
        """

        average_holding = average_holding_period(self.hold_positions)
        nan_average_holding = average_holding_period(self.no_closed_positions)

        # As seen from example set, positions are kept 2 days on average
        self.assertAlmostEqual(average_holding, 2, delta=1e-4)
        self.assertTrue(np.isnan(nan_average_holding))

    def test_bets_concentration(self):
        """
        Check if concentration is balanced and correctly calculated
        """

        positive_concentration = bets_concentration(self.logret)
        #Testing for symmetry in concentration
        flipped_logret = -1 * self.logret
        negative_concentration = bets_concentration(flipped_logret)

        self.assertAlmostEqual(positive_concentration, negative_concentration,
                               delta=1e-5)
        self.assertAlmostEqual(positive_concentration, 2.0111445, delta=1e-4)

    def test_all_bets_concentration(self):
        """
        Check if concentration is nan when not enough observations, also values
        testing
        """

        # Only one negative return in this dataset
        positive_returns_concentration = all_bets_concentration(self.normal_returns)

        # A longer dataset with all concentrations
        all_returns_concentration = all_bets_concentration(self.logret, frequency='D')

        # Not enough negative observations in dataset
        self.assertTrue(np.isnan(positive_returns_concentration[1]))

        # Not enough observations to group by month
        self.assertTrue(np.isnan(positive_returns_concentration[2]))
        self.assertAlmostEqual(all_returns_concentration[0], 0.0014938,
                               delta=1e-5)
        self.assertAlmostEqual(all_returns_concentration[1], 0.0016261,
                               delta=1e-5)
        self.assertAlmostEqual(all_returns_concentration[2], 0.0195998,
                               delta=1e-5)

    def test_drawdown_and_time_under_water(self):
        """
        Check if drawdowns and time under water calculated correctly for
        dollar and non-dollar test sets.
        """

        drawdown_dol, time_under_water_dol = drawdown_and_time_under_water(self.dollar_returns,
                                                                           dollars=True)
        _, time_under_water = drawdown_and_time_under_water(self.dollar_returns / 100,
                                                            dollars=False)

        self.assertTrue(list(drawdown_dol) == [20.0, 30.0, 10.0])
        self.assertTrue(list(time_under_water) == list(time_under_water_dol))
        self.assertAlmostEqual(time_under_water[0], 0.010951,
                               delta=1e-4)
        self.assertAlmostEqual(time_under_water[1], 0.008213,
                               delta=1e-4)

    def test_sharpe_ratio(self):
        """
        Check if Sharpe ratio is calculated right
        """

        sharpe = sharpe_ratio(self.normal_returns, entries_per_year=12,
                              risk_free_rate=0.005)

        self.assertAlmostEqual(sharpe, 0.987483, delta=1e-4)

    def test_information_ratio(self):
        """
        Check if Information ratio is calculated right
        """

        information_r = information_ratio(self.normal_returns, benchmark=0.006,
                                          entries_per_year=12)

        self.assertAlmostEqual(information_r, 0.733559, delta=1e-4)

    def test_probabilistic_sharpe_ratio(self):
        """
        Check probabilistic Sharpe ratio using numerical example
        """

        observed_sr = 1.14
        benchmark_sr = 1
        number_of_returns = 250
        skewness = 0
        kurtosis = 3

        result_prob_sr = probabilistic_sharpe_ratio(observed_sr, benchmark_sr,
                                                    number_of_returns, skewness,
                                                    kurtosis)

        self.assertAlmostEqual(result_prob_sr, 0.95727,
                               delta=1e-4)

        # Test warnings

        with self.assertWarns(UserWarning):  # Test statistic is complex
            probabilistic_sharpe_ratio(1.2, 1.0, 2, 2, 3)

        with self.assertWarns(UserWarning):  # Test statistic is inf
            probabilistic_sharpe_ratio(2, 1.0, 2, 0.5, 1)

        with self.assertWarns(UserWarning):  # Test statistic is nan
            probabilistic_sharpe_ratio(2, 1.0, 1, 0.5, 1)

    def test_deflated_sharpe_ratio(self):
        """
        Check deflated Sharpe ratio using numerical example
        """

        observed_sr = 1.14
        sr_estimates = [3.5, 1.01, 1.02]

        # Parameters of SR estimates - standard deviation and number of observations
        estim_param = [0.4, 100]
        number_of_returns = 250
        skewness = 0
        kurtosis = 3

        result_defl_sr = deflated_sharpe_ratio(observed_sr, sr_estimates,
                                               number_of_returns, skewness,
                                               kurtosis)

        benchmark_sr = deflated_sharpe_ratio(observed_sr, estim_param,
                                             number_of_returns, skewness,
                                             kurtosis, estimates_param=True,
                                             benchmark_out=True)

        param_defl_sr = deflated_sharpe_ratio(observed_sr, estim_param,
                                              number_of_returns, skewness,
                                              kurtosis, estimates_param=True)


        self.assertAlmostEqual(result_defl_sr, 0.95836,
                               delta=1e-4)

        self.assertAlmostEqual(benchmark_sr, 1.012241,
                               delta=1e-4)

        self.assertAlmostEqual(param_defl_sr, 0.941740,
                               delta=1e-4)

    def test_minimum_track_record_length(self):
        """
        Check deflated Sharpe ratio using numerical example
        """

        observed_sr = 1.14
        benchmark_sr = 1
        skewness = 0
        kurtosis = 3
        alpha = 0.05

        result_min_track_rec = minimum_track_record_length(observed_sr,
                                                           benchmark_sr,
                                                           skewness,
                                                           kurtosis,
                                                           alpha)

        self.assertAlmostEqual(result_min_track_rec, 228.73497,
                               delta=1e-4)



"""
This implementation lets user generate dependence and distance matrix based on the various methods of Information
Codependence  described in Cornell lecture notes on Codependence:
https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes
"""

import numpy as np
import pandas as pd

from famlafl.codependence.information import variation_of_information_score, get_mutual_info
from famlafl.codependence.correlation import distance_correlation
from famlafl.codependence.gnpr_distance import spearmans_rho, gpr_distance, gnpr_distance

# pylint: disable=invalid-name

def get_dependence_matrix(df: pd.DataFrame, dependence_method: str, theta: float = 0.5,
                          bandwidth: float = 0.01) -> pd.DataFrame:
    """
    This function returns a dependence matrix for elements given in the dataframe using the chosen dependence method.

    List of supported algorithms to use for generating the dependence matrix: ``information_variation``,
    ``mutual_information``, ``distance_correlation``, ``spearmans_rho``, ``gpr_distance``, ``gnpr_distance``.

    :param df: (pd.DataFrame) Features.
    :param dependence_method: (str) Algorithm to be use for generating dependence_matrix.
    :param theta: (float) Type of information being tested in the GPR and GNPR distances. Falls in range [0, 1].
                          (0.5 by default)
    :param bandwidth: (float) Bandwidth to use for splitting observations in the GPR and GNPR distances. (0.01 by default)
    :return: (pd.DataFrame) Dependence matrix.
    """
    # Get the feature names.
    features_cols = df.columns.values
    n = df.shape[1]
    np_df = df.values.T  # Make columnar access, but for np.array

    # Defining the dependence function.
    if dependence_method == 'information_variation':
        dep_function = lambda x, y: variation_of_information_score(x, y, normalize=True)
    elif dependence_method == 'mutual_information':
        dep_function = lambda x, y: get_mutual_info(x, y, normalize=True)
    elif dependence_method == 'distance_correlation':
        dep_function = distance_correlation
    elif dependence_method == 'spearmans_rho':
        dep_function = spearmans_rho
    elif dependence_method == 'gpr_distance':
        dep_function = lambda x, y: gpr_distance(x, y, theta=theta)
    elif dependence_method == 'gnpr_distance':
        dep_function = lambda x, y: gnpr_distance(x, y, theta=theta, bandwidth=bandwidth)
    else:
        raise ValueError(f"{dependence_method} is not a valid method. Please use one of the supported methods \
                            listed in the docsting.")

    # Generating the dependence_matrix
    dependence_matrix = np.array([
        [
            dep_function(np_df[i], np_df[j]) if j < i else
            # Leave diagonal elements as 0.5 to later double them to 1
            0.5 * dep_function(np_df[i], np_df[j]) if j == i else
            0  # Make upper triangle 0 to fill it later on
            for j in range(n)
        ]
        for i in range(n)
    ])

    # Make matrix symmetrical
    dependence_matrix = dependence_matrix + dependence_matrix.T

    #  Dependence_matrix converted into a DataFrame.
    dependence_df = pd.DataFrame(data=dependence_matrix, index=features_cols, columns=features_cols)

    if dependence_method == 'information_variation':
        return 1 - dependence_df  # IV is reverse, 1 - independent, 0 - similar

    return dependence_df

def get_distance_matrix(X: pd.DataFrame, distance_metric: str = 'angular') -> pd.DataFrame:
    """
    Applies distance operator to a dependence matrix.

    This allows to turn a correlation matrix into a distance matrix. Distances used are true metrics.

    List of supported distance metrics to use for generating the distance matrix: ``angular``, ``squared_angular``,
    and ``absolute_angular``.

    :param X: (pd.DataFrame) Dataframe to which distance operator to be applied.
    :param distance_metric: (str) The distance metric to be used for generating the distance matrix.
    :return: (pd.DataFrame) Distance matrix.
    """
    if distance_metric == 'angular':
        distfun = lambda x: ((1 - x).round(5) / 2.) ** .5
    elif distance_metric == 'abs_angular':
        distfun = lambda x: ((1 - abs(x)).round(5) / 2.) ** .5
    elif distance_metric == 'squared_angular':
        distfun = lambda x: ((1 - x ** 2).round(5) / 2.) ** .5
    else:
        raise ValueError(f'{distance_metric} is a unknown distance metric. Please use one of the supported methods \
                            listed in the docsting.')

    return distfun(X).fillna(0)



"""
Tests Constant Rebalanced Portfolio.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection import CRP


class TestConstantRebalancedPortfolio(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the CRP class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_default_crp_solution(self):
        """
        Tests the calculation of CRP weights with default settings.
        """
        # Initialize CRP.
        crp = CRP()
        # Allocates asset prices to CRP.
        crp.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(crp.all_weights)
        # All weights for the strategy have to be the same.
        one_weight = all_weights[0]
        # Iterate through all_weights to check that weights equal to the first weight.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            assert (weights == one_weight).all()
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_given_weights_crp_solution(self):
        """
        Tests the calculation of constant rebalanced portfolio weights with weights given initially.
        """
        # Create user input weights
        weights = np.zeros(self.data.shape[1])
        # Set 1 on the first stock and 0 on the rest.
        weights[0] = 1
        # Initialize CRP.
        crp = CRP(weights)
        # Allocates asset prices to CRP.
        crp.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(crp.all_weights)
        # All weights for the strategy have to be the same.
        one_weight = all_weights[0]
        # Iterate through all_weights to check that weights equal to the first weight.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            assert (weights == one_weight).all()
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_given_allocate_weights_crp_solution(self):
        """
        Test calculation of constant rebalanced portfolio weights with weights given in allocate.
        """
        # Create user input weights
        weights = np.zeros(self.data.shape[1])
        # Set 1 on the first stock and 0 on the rest.
        weights[0] = 1
        # Initialize CRP.
        crp = CRP()
        # Allocates asset prices to CRP.
        crp.allocate(self.data, weights, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(crp.all_weights)
        # All weights for the strategy have to be the same.
        one_weight = all_weights[0]
        # Iterate through all_weights to check that weights equal to the first weight.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            assert (weights == one_weight).all()
            np.testing.assert_almost_equal(np.sum(weights), 1)



"""
Optimal Number of Clusters (ONC Algorithm)
Detection of False Investment Strategies using Unsupervised Learning Methods
https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3167017
"""

from typing import Union

import numpy as np
import pandas as pd

from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_samples


def _improve_clusters(corr_mat: pd.DataFrame, clusters: dict, top_clusters: dict) -> Union[
        pd.DataFrame, dict, pd.Series]:
    """
    Improve number clusters using silh scores

    :param corr_mat: (pd.DataFrame) Correlation matrix
    :param clusters: (dict) Clusters elements
    :param top_clusters: (dict) Improved clusters elements
    :return: (tuple) [ordered correlation matrix, clusters, silh scores]
    """
    clusters_new, new_idx = {}, []
    for i in clusters.keys():
        clusters_new[len(clusters_new.keys())] = list(clusters[i])

    for i in top_clusters.keys():
        clusters_new[len(clusters_new.keys())] = list(top_clusters[i])

    map(new_idx.extend, clusters_new.values())
    corr_new = corr_mat.loc[new_idx, new_idx]

    dist = ((1 - corr_mat.fillna(0)) / 2.0) ** 0.5

    kmeans_labels = np.zeros(len(dist.columns))
    for i in clusters_new:
        idxs = [dist.index.get_loc(k) for k in clusters_new[i]]
        kmeans_labels[idxs] = i

    silh_scores_new = pd.Series(silhouette_samples(dist, kmeans_labels), index=dist.index)
    return corr_new, clusters_new, silh_scores_new


def _cluster_kmeans_base(corr_mat: pd.DataFrame, max_num_clusters: int = 10, repeat: int = 10) -> Union[
        pd.DataFrame, dict, pd.Series]:
    """
    Initial clustering step using KMeans.

    :param corr_mat: (pd.DataFrame) Correlation matrix
    :param max_num_clusters: (int) Maximum number of clusters to search for.
    :param repeat: (int) Number of clustering algorithm repetitions.
    :return: (tuple) [ordered correlation matrix, clusters, silh scores]
    """

    # Distance matrix

    # Fill main diagonal of corr matrix with 1s to avoid elements being close to 1 with e-16.
    # As this previously caused Errors when taking square root from negative values.
    np.fill_diagonal(corr_mat.values, 1)
    distance = ((1 - corr_mat.fillna(0)) / 2.0) ** 0.5
    silh = pd.Series(dtype='float64')

    # Get optimal num clusters
    for _ in range(repeat):
        for num_clusters in range(2, max_num_clusters + 1):
            kmeans_ = KMeans(n_clusters=num_clusters, n_init=1)
            kmeans_ = kmeans_.fit(distance)
            silh_ = silhouette_samples(distance, kmeans_.labels_)
            stat = (silh_.mean() / silh_.std(), silh.mean() / silh.std())

            if np.isnan(stat[1]) or stat[0] > stat[1]:
                silh = silh_
                kmeans = kmeans_

    # Number of clusters equals to length(kmeans labels)
    new_idx = np.argsort(kmeans.labels_)

    # Reorder rows
    corr1 = corr_mat.iloc[new_idx]
    # Reorder columns
    corr1 = corr1.iloc[:, new_idx]

    # Cluster members
    clusters = {i: corr_mat.columns[np.where(kmeans.labels_ == i)[0]].tolist() for i in
                np.unique(kmeans.labels_)}
    silh = pd.Series(silh, index=distance.index)

    return corr1, clusters, silh


def _check_improve_clusters(new_tstat_mean: float, mean_redo_tstat: float, old_cluster: tuple,
                            new_cluster: tuple) -> tuple:
    """
    Checks cluster improvement condition based on t-statistic.

    :param new_tstat_mean: (float) T-statistics
    :param mean_redo_tstat: (float) Average t-statistcs for cluster improvement
    :param old_cluster: (tuple) Old cluster correlation matrix, optimized clusters, silh scores
    :param new_cluster: (tuple) New cluster correlation matrix, optimized clusters, silh scores
    :return: (tuple) Cluster
    """

    if new_tstat_mean > mean_redo_tstat:
        return old_cluster
    return new_cluster


def cluster_kmeans_top(corr_mat: pd.DataFrame, repeat: int = 10) -> Union[pd.DataFrame, dict, pd.Series, bool]:
    """
    Improve the initial clustering by leaving clusters with high scores unchanged and modifying clusters with
    below average scores.

    :param corr_mat: (pd.DataFrame) Correlation matrix
    :param repeat: (int) Number of clustering algorithm repetitions.
    :return: (tuple) [correlation matrix, optimized clusters, silh scores, boolean to rerun ONC]
    """
    # pylint: disable=no-else-return

    max_num_clusters = min(corr_mat.drop_duplicates().shape[0], corr_mat.drop_duplicates().shape[1]) - 1
    corr1, clusters, silh = _cluster_kmeans_base(corr_mat, max_num_clusters=max_num_clusters, repeat=repeat)

    # Get cluster quality scores
    cluster_quality = {i: float('Inf') if np.std(silh[clusters[i]]) == 0 else np.mean(silh[clusters[i]]) /
                          np.std(silh[clusters[i]]) for i in clusters.keys()}
    avg_quality = np.mean(list(cluster_quality.values()))
    redo_clusters = [i for i in cluster_quality.keys() if cluster_quality[i] < avg_quality]

    if len(redo_clusters) <= 2:
        # If 2 or less clusters have a quality rating less than the average then stop.
        return corr1, clusters, silh
    else:
        keys_redo = []
        for i in redo_clusters:
            keys_redo.extend(clusters[i])

        corr_tmp = corr_mat.loc[keys_redo, keys_redo]
        mean_redo_tstat = np.mean([cluster_quality[i] for i in redo_clusters])
        _, top_clusters, _ = cluster_kmeans_top(corr_tmp, repeat=repeat)

        # Make new clusters (improved)
        corr_new, clusters_new, silh_new = _improve_clusters(corr_mat,
                                                             {i: clusters[i] for i in clusters.keys() if
                                                              i not in redo_clusters},
                                                             top_clusters)
        new_tstat_mean = np.mean(
            [np.mean(silh_new[clusters_new[i]]) / np.std(silh_new[clusters_new[i]]) for i in clusters_new])

        return _check_improve_clusters(new_tstat_mean, mean_redo_tstat, (corr1, clusters, silh),
                                       (corr_new, clusters_new, silh_new))


def get_onc_clusters(corr_mat: pd.DataFrame, repeat: int = 10) -> Union[pd.DataFrame, dict, pd.Series]:
    """
    Optimal Number of Clusters (ONC) algorithm described in the following paper:
    `Marcos Lopez de Prado, Michael J. Lewis, Detection of False Investment Strategies Using Unsupervised
    Learning Methods, 2015 <https://papers.ssrn.com/sol3/abstract_id=3167017>`_;
    The code is based on the code provided by the authors of the paper.

    The algorithm searches for the optimal number of clusters using the correlation matrix of elements as an input.

    The correlation matrix is transformed to a matrix of distances, the K-Means algorithm is applied multiple times
    with a different number of clusters to use. The results are evaluated on the t-statistics of the silhouette scores.

    The output of the algorithm is the reordered correlation matrix (clustered elements are placed close to each other),
    optimal clustering, and silhouette scores.

    :param corr_mat: (pd.DataFrame) Correlation matrix of features
    :param repeat: (int) Number of clustering algorithm repetitions
    :return: (tuple) [correlation matrix, optimized clusters, silh scores]
    """

    return cluster_kmeans_top(corr_mat, repeat)



"""
Tests the cross validation technique described in Ch.7 of the book.
"""
import unittest
import pandas as pd
from famlafl.cross_validation import CombinatorialPurgedKFold


class TestCombinatorialPurgedCV(unittest.TestCase):
    """
    Test the functionality of the time series cross validation technique
    """

    def setUp(self):
        """
        Test Combinatorial Purged CV class
        """

        self.info_sets = pd.Series(
            index=pd.date_range(start='2019-01-01 00:00:00', periods=100, freq='D'),
            data=pd.date_range(start='2019-01-02 00:02:00', periods=100, freq='T'),
        )

    def test_test_times(self):
        """
        Tests the Combinatorial CV with various number of train, test splits
        """
        cv_gen = CombinatorialPurgedKFold(n_splits=6, n_test_splits=2, samples_info_sets=self.info_sets)
        counter = 0  # number of splits counter
        train_splits = []  # Array of train indices splits
        test_splits = []  # Array of test indices splits

        for train, test in cv_gen.split(self.info_sets):
            counter += 1
            train_splits.append(train)
            test_splits.append(test)

        self.assertEqual(counter, 15)  # Example from the book says that for CombinatorialCV(6, 2) number of splits=15

        # Check indices from example from the book (page 164)
        for test_idx, [start_idx, end_idx] in zip(test_splits[0:5], [[0, 33], [0, 50], [0, 67], [0, 83], [0, 99]]):
            self.assertEqual(test_idx[0], start_idx)
            self.assertEqual(test_idx[-1], end_idx)

        for test_idx, [start_idx, end_idx] in zip(test_splits[-5:], [[34, 83], [34, 99], [51, 83], [51, 99], [68, 99]]):
            self.assertEqual(test_idx[0], start_idx)
            self.assertEqual(test_idx[-1], end_idx)

        # Test backtest paths
        self.assertEqual(len(cv_gen.backtest_paths), cv_gen.n_splits - 1)  # phi[N, 2] = N - 1
        self.assertTrue((cv_gen.backtest_paths[0][0]['train'] == train_splits[0]).all())
        self.assertTrue((cv_gen.backtest_paths[0][1]['train'] == train_splits[0]).all())
        self.assertTrue((cv_gen.backtest_paths[1][0]['train'] == train_splits[1]).all())
        self.assertTrue((cv_gen.backtest_paths[1][2]['train'] == train_splits[5]).all())
        self.assertTrue((cv_gen.backtest_paths[-1][-1]['train'] == train_splits[-1]).all())
        self.assertTrue((cv_gen.backtest_paths[-1][-2]['train'] == train_splits[-1]).all())


    def test_purge_and_embargo(self):
        """
        Test that purging and embargo works correctly
        """

        cv_gen = CombinatorialPurgedKFold(n_splits=6, n_test_splits=2, samples_info_sets=self.info_sets,
                                          pct_embargo=0.05)
        train_splits = []
        test_splits = []

        for train, test in cv_gen.split(self.info_sets):
            train_splits.append(train)
            test_splits.append(test)

        cv_gen_no_embargo = CombinatorialPurgedKFold(n_splits=6, n_test_splits=2, samples_info_sets=self.info_sets,
                                                     pct_embargo=0.0)
        train_splits_no = []
        test_splits_no = []

        for train, test in cv_gen_no_embargo.split(self.info_sets):
            train_splits_no.append(train)
            test_splits_no.append(test)

        self.assertEqual(len(train_splits_no[0]) - len(train_splits[0]), 5)  # Embargo of 1 train set
        self.assertEqual(len(train_splits_no[1]) - len(train_splits[1]), 10)  # Embargo of 2 train set
        self.assertEqual(len(train_splits_no[-1]) - len(train_splits[-1]), 0)  # Latest set is not embargo-ed

    def test_errors_raise(self):
        """
        Test if ValueErrors are raised
        """

        with self.assertRaises(ValueError):
            CombinatorialPurgedKFold(n_splits=6, n_test_splits=3, samples_info_sets=[1, 2, 3, 4])

        cv_gen = CombinatorialPurgedKFold(n_splits=6, n_test_splits=2, samples_info_sets=self.info_sets,
                                          pct_embargo=0.05)

        with self.assertRaises(ValueError):
            for _, _ in cv_gen.split(self.info_sets.iloc[2:]):
                pass



"""
Functions derived from Chapter 19: Market Microstructural features
"""

from famlafl.microstructural_features.encoding import quantile_mapping, encode_array, encode_tick_rule_array, \
    sigma_mapping
from famlafl.microstructural_features.entropy import get_lempel_ziv_entropy, get_shannon_entropy, get_plug_in_entropy, \
    get_konto_entropy
from famlafl.microstructural_features.feature_generator import MicrostructuralFeaturesGenerator
from famlafl.microstructural_features.first_generation import get_corwin_schultz_estimator, get_roll_measure, \
    get_roll_impact, get_bekker_parkinson_vol
from famlafl.microstructural_features.misc import get_avg_tick_size, vwap
from famlafl.microstructural_features.second_generation import get_bar_based_kyle_lambda, get_bar_based_amihud_lambda, \
    get_bar_based_hasbrouck_lambda, get_trades_based_kyle_lambda, get_trades_based_amihud_lambda, \
    get_trades_based_hasbrouck_lambda
from famlafl.microstructural_features.third_generation import get_vpin



"""
The module implementing various functions loading tick, dollar, stock data sets which can be used as
sandbox data
"""

import os
import pandas as pd


def load_stock_prices() -> pd.DataFrame:
    """
    Loads stock prices data sets consisting of
    EEM, EWG, TIP, EWJ, EFA, IEF, EWQ, EWU, XLB, XLE, XLF, LQD, XLK, XLU, EPP, FXI, VGK, VPL, SPY, TLT, BND, CSJ,
    DIA starting from 2008 till 2016.

    :return: (pd.DataFrame) stock_prices data frame
    """

    project_path = os.path.dirname(__file__)
    prices_df = pd.read_csv(os.path.join(project_path, 'data/stock_prices.csv'), index_col=0, parse_dates=[0])
    return prices_df


def load_tick_sample() -> pd.DataFrame:
    """
    Loads E-Mini S&P 500 futures tick data sample

    :return: (pd.DataFrame) with tick data sample
    """

    project_path = os.path.dirname(__file__)
    tick_df = pd.read_csv(os.path.join(project_path, 'data/tick_data.csv'), index_col=0, parse_dates=[0])
    return tick_df


def load_dollar_bar_sample() -> pd.DataFrame:
    """
    Loads E-Mini S&P 500 futures dollar bars data sample.

    :return: (pd.DataFrame) with dollar bar data sample
    """

    project_path = os.path.dirname(__file__)
    bars_df = pd.read_csv(os.path.join(project_path, 'data/dollar_bar_sample.csv'), index_col=0, parse_dates=[0])
    return bars_df



# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd


class TailSetLabels:
    """
    Tail set labels are a classification labeling technique introduced in the following paper: Nonlinear support vector
    machines can systematically identify stocks with high and low future returns. Algorithmic Finance, 2(1), pp.45-58.

    A tail set is defined to be a group of stocks whose volatility-adjusted return is in the highest or lowest
    quantile, for example the highest or lowest 5%.

    A classification model is then fit using these labels to determine which stocks to buy and sell in a long / short
    portfolio.
    """

    def __init__(self, prices, n_bins, vol_adj=None, window=None):
        """
        :param prices: (pd.DataFrame) Asset prices.
        :param n_bins: (int) Number of bins to determine the quantiles for defining the tail sets. The top and
                        bottom quantiles are considered to be the positive and negative tail sets, respectively.
        :param vol_adj: (str) Whether to take volatility adjusted returns. Allowable inputs are ``None``,
                        ``mean_abs_dev``, and ``stdev``.
        :param window: (int) Window period used in the calculation of the volatility adjusted returns, if vol_adj is not
                        None. Has no impact if vol_adj is None.
        """
        assert prices.shape[1] >= n_bins, "n_bins exceeds the number of stocks!"
        if vol_adj is not None:
            assert isinstance(window, int), "If vol_adj is not None, window must be int."
            assert len(prices) > window, "Length of price data must be greater than the window."

        self.prices = prices
        self.rets = np.log(prices).diff().dropna()
        self.window = window
        self.n_bins = n_bins
        self.vol_adj = vol_adj

        # Properties relating to the tail sets.
        self.vol_adj_rets = self.rets  # If no vol_adj.

        # Compute tail sets.
        if self.vol_adj is not None:
            self._vol_adjusted_rets()
        self.tail_sets = self.vol_adj_rets.dropna().apply(self._extract_tail_sets, axis=1)
        self.positive_tail_set = self.tail_sets.apply(self._positive_tail_set, axis=1)
        self.negative_tail_set = self.tail_sets.apply(self._negative_tail_set, axis=1)

    def get_tail_sets(self):
        """
        Computes the tail sets (positive and negative) and then returns a tuple with 3 elements, positive set, negative
        set, full matrix set.

        The positive and negative sets are each a series of lists with the names of the securities that fall within each
        set at a specific timestamp.

        For the full matrix a value of 1 indicates the volatility adjusted returns were in the top quantile, a value of
        -1 for the bottom quantile.
        :return: (tuple) positive set, negative set, full matrix set.
        """
        return self.positive_tail_set, self.negative_tail_set, self.tail_sets

    def _vol_adjusted_rets(self):
        """
        Computes the volatility adjusted returns. This is simply the log returns divided by a volatility estimate. We
        have provided 2 techniques for volatility estimation: an exponential moving average and the traditional standard
        deviation.
        """
        window = self.window

        # Have 2 measure of vol, the mean absolute, and stdev.
        if self.vol_adj == 'mean_abs_dev':
            # Huffman and Moll (2011) show that risk measured as the mean absolute deviation has more explanatory
            # power for future expected returns than standard deviation.
            vol = self.rets.abs().ewm(span=window, min_periods=window).mean()
        elif self.vol_adj == 'stdev':
            vol = self.rets.rolling(window).std()
        else:
            raise Exception('Invalid name for vol_adj. Valid names are ''mean_abs_dev'', ''stdev'', or None.')
        # Save vol adj rets.
        self.vol_adj_rets = (self.rets / vol).dropna()

    def _extract_tail_sets(self, row):
        """
        Method used in a .apply() setting to transform each row in a DataFrame to the positive and negative tail sets.

        This method splits the data into quantiles determined by the user, with n_bins.

        :param row: (pd.Series) Vol adjusted returns for a given date.
        :return: (pd.Series) Tail set with positive and negative labels.
        """
        # Get quantile labels.
        row = row.rank(method='first')  # To avoid error with unique bins when using qcut due to too many 0 values.
        row_quantiles = pd.qcut(x=row, q=self.n_bins, labels=range(1, 1 + self.n_bins), retbins=False)

        # Set class labels.
        row_quantiles = row_quantiles.to_numpy()  # Convert to numpy array
        row_quantiles[(row_quantiles != 1) & (row_quantiles != self.n_bins)] = 0
        row_quantiles[row_quantiles == 1] = -1
        row_quantiles[row_quantiles == self.n_bins] = 1

        # Convert to series.
        row_quantiles = pd.Series(row_quantiles, index=row.index)

        return row_quantiles

    @staticmethod
    def _positive_tail_set(row):
        """
        Takes as input a row from the vol_adj_ret DataFrame and then returns a list of names of the securities in the
        positive tail set, for this specific row date.

        This method is used in an apply() setting.

        :param row: (pd.Series) Labeled row of several stocks where each is already labeled with +1 (positive tail set),
                    -1 (negative tail set), or 0.
        :return: (list) Securities in the positive tail set.
        """
        return list(row[row == 1].index)

    @staticmethod
    def _negative_tail_set(row):
        """
        Takes as input a row from the vol_adj_ret DataFrame and then returns a list of names of the securities in the
        negative tail set, for this specific row date.

        This method is used in an apply() setting.

        :param row: (pd.Series) Labeled row of several stocks where each is already labeled with +1 (positive tail set),
                    -1 (negative tail set), or 0.
        :return: (list) Securities in the negative tail set.
        """
        return list(row[row == -1].index)



# pylint: disable=missing-module-docstring

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.labeling.excess_over_mean import excess_over_mean


class TestLabelingOverMean(unittest.TestCase):
    """
    Tests regarding labeling excess over median.
    """

    def setUp(self):
        """
        Set the file path for the sample data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(self.path, index_col='Date')
        self.data.index = pd.to_datetime(self.data.index)

    def test_small_set(self):
        """
        Check for a small set with manually inputted results, with numerical and categorical outputs.
        """
        cols = ['EEM', 'EWG', 'TIP', 'EWJ']
        subset = self.data[cols].iloc[0:5]
        test1 = excess_over_mean(subset)
        test2 = excess_over_mean(subset, binary=True)
        test1_actual = pd.DataFrame([(0.005666, -0.006157, 0.00004411, 0.0004476),
                                     (-0.011169, -0.000684, 0.018588, -0.006734),
                                     (0.006871, 0.000411, -0.000643, -0.006639),
                                     (-0.003687, -0.002863, 0.004815, 0.001735),
                                     (np.nan, np.nan, np.nan, np.nan), ], columns=self.data[cols].iloc[0:5].columns,
                                    index=self.data[cols].iloc[0:5].index)
        test2_actual = test1_actual.apply(np.sign)

        pd.testing.assert_frame_equal(test1, test1_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test2, test2_actual)

    def test_large_set(self):
        """
        Checks a specific row for a large dataset, and ensures the last row is NaN.
        """
        # Get a random row of the entire dataset.
        test3 = pd.Series(excess_over_mean(self.data, lag=True).iloc[42])
        idx42 = self.data.iloc[42].index
        test3_actual = pd.Series([0.014267, 0.00537, -0.007375, -0.001013, 0.005282, -0.011075, 0.006921, 0.000626,
                                  0.021376, 0.018429, -0.01463, -0.01377, 0.00048, -0.009569, 0.0107, 0.0121335,
                                  0.0050173, 0.0002504, -0.0004207, -0.018707, -0.013208, -0.00818807, -0.002897],
                                 index=idx42)

        pd.testing.assert_series_equal(test3, test3_actual, rtol=0.001, check_names=False)

        test4 = excess_over_mean(self.data, lag=True).iloc[-1]
        self.assertTrue(test4.isnull().all())

    def test_resample_period(self):
        """
        Test numerical and categorical labels with a resample period.
        """
        cols = ['EEM', 'EWG', 'TIP', 'EWJ']
        subset1 = self.data[cols].iloc[0:25]
        subset2 = self.data[cols].iloc[0:100]
        week_idx = subset1.resample('W').last().index
        month_idx = subset2.resample('M').last().index
        # Resample per week and per month.
        test5 = excess_over_mean(subset1, binary=False, resample_by='W', lag=True)
        test6 = excess_over_mean(subset2, binary=False, resample_by='M', lag=False)
        test7 = excess_over_mean(subset2, binary=True, resample_by='M', lag=False)

        test5_actual = pd.DataFrame({'EEM': [0.017255, -0.042112, 0.004907, 0.016267, -0.026054, np.nan],
                                     'EWG': [-0.011975, -0.019257, -0.048906, 0.017310, -0.003467, np.nan],
                                     'TIP': [0.004898, 0.047419, 0.026425, -0.031773, 0.038212, np.nan],
                                     'EWJ': [-0.010178, 0.013950, 0.017574, -0.001804, -0.008691, np.nan]},
                                    index=week_idx)
        test6_actual = pd.DataFrame({'EEM': [np.nan, 0.018196, -0.027718, 0.047210, 0.001358],
                                     'EWG': [np.nan, -0.008799, 0.026921, -0.007042, 0.007712],
                                     'TIP': [np.nan, 0.0072872, 0.0037534, -0.069560, 0.0004006],
                                     'EWJ': [np.nan, -0.016684, -0.002956, 0.029392, -0.009471]},
                                    index=month_idx)

        pd.testing.assert_frame_equal(test5, test5_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test6, test6_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test7, test6_actual.apply(np.sign))



"""
Tests Exponential Gradient.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.eg import EG


class TestExponentialGradient(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Exponential Gradient class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_mu_solution(self):
        """
        Test calculation of exponential gradient weights with multiplicative update rule.
        """
        # Use multiplicative update rule.
        multiplicative_update = EG(update_rule='MU')
        # Allocates asset prices to MU.
        multiplicative_update.allocate(self.data, resample_by='M')
        all_weights = np.array(multiplicative_update.all_weights)
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_gp_solution(self):
        """
        Test calculation of exponential gradient weights with gradient projection update rule.
        """
        # Use gradient projection update rule.
        gradient_projection = EG(update_rule='GP')
        # Allocates asset prices to GP.
        gradient_projection.allocate(self.data, resample_by='M')
        all_weights = np.array(gradient_projection.all_weights)
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_em_solution(self):
        """
        Test calculation of exponential gradient weights with expectation maximization update rule.
        """
        # Use expectation maximization update rule.
        expectation_maximization = EG(update_rule='EM')
        # Allocates asset prices to EM.
        expectation_maximization.allocate(self.data, resample_by='M')
        all_weights = np.array(expectation_maximization.all_weights)
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_wrong_update(self):
        """
        Tests ValueError if the passing update rule is not correct.
        """
        # Put in incorrect update rule.
        expectation_maximization = EG(update_rule='SS')
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            expectation_maximization.allocate(self.data, resample_by='M')
